package modules;

import java.io.IOException;
import java.util.Random;
import java.io.File;
import java.net.InetAddress;

import types.*;
import modules.DebugCollector.DebugCollectorObserver;
import modules.ErrorCollector.ErrorCollectorObserver;

/*
 * peer lists
 * shared files
 * search results 
 * download descriptors
 * streamID-s
 * ...
 */

/** === DataStorage class ===                         **/
/** This class manages the databases                  **/
/**                                                   **/
/** !!! NOTE: This class is a singleton !!!           **/
/** Exactly one object is expected to be instantiated **/
//Developer Note (Peto) 
//--Singleton is used to be implemented with protected constructor 
//--and public static getInstance() - it ensures that just one instance is made.
//--Also then the reference is stored at one place so it don`t have to be stored more times.
//Remark (Imp) - and an exception throwing if clone() is called... I see your point,
//but find it unnecessary in this case; most of the classes in this project are singletons, anyway.

public class DataStorage{
    /** GENERAL MEMBERS **/

		private boolean debugActive = false;
	
		// modules
		protected Communication communication;
		protected SearchModul searchModule;
		protected SearchAnswerModul searchAnswerModule;
		protected DownloadModul downloadModul;
		protected UploadModul uploadModul;
		protected PeerManagementModul peerManagementModul;
	
        // internal database
		private static int DEFAULT_PORT = 4747;
		private static int port; 
		private static boolean portSet = false;
        
		private int myPort;
        private PeerList peerList;
        private IDGenerator streamIDGenerator;
        private PeerID myPeerID;
        private SearchResultList resultList;
        private ErrorCollector errorCollector;
        private DebugCollector debugCollector;
        private boolean lastPeerListBackup;
        private static final String peerFile1 = "peerlistA.txt";
        private static final String peerFile2 = "peerlistB.txt";
        private Random random;
        private File configDir;
        private RecentlyDeletedList rdList;
        
        //Stores DownloadedFileDescriptors for files in progress of download
        private DownloadedFilesDatabase down_files_db; 
    	public static final String down_db_init_file = "down.ini";
        
        //Stores SharedFilesDescriptors
        private SharedFilesDatabase shared_files_db;
    	public static final String shared_db_ini_file = "shared.ini";
        
        //
        private ProgressSignalReceiver currentProgressSignalReceiver = null;
        
        // constructor
        public DataStorage() {
                myPort = portSet ? port : DEFAULT_PORT;
                
                peerList = new PeerList();
                streamIDGenerator = new IDGenerator();
                random = new Random();
                myPeerID = new PeerID(random.nextInt());
                resultList = new SearchResultList();
                lastPeerListBackup = false;
                
                configDir = new File(getClass().getResource(".").getFile() + "../config/");
                if (!configDir.mkdir() && !configDir.exists()) {
                        System.out.println("*** Error creating the configuration directory. ***");
                }
                
                try {
                        new File(configPath() + peerFile1).createNewFile();
                        new File(configPath() + peerFile2).createNewFile();
                } catch (IOException e) {
                        System.out.println("*** Error creating the peerlist backup files. ***");
                }
                loadPeerListBackup();
                
                this.errorCollector = new ErrorCollector();
                this.debugCollector = new DebugCollector();

        //!TODO
        //Developer Note (Peto): I am not experienced much in multithreaded apps.
        //But what happens if one thread is iterating through the set and the another modifies it ?
        //Answer (Imp): Bad things happen. First, create the set as synchronized, e.g.:
        //mySet = Collection.synchronizedSortedSet(new TreeSet<...>); Then, use synchronized on
        //all the important method calls
                
                this.down_files_db = new DownloadedFilesDatabase();
                this.shared_files_db = new SharedFilesDatabase();
                
                rdList = new RecentlyDeletedList();
        }
    
    /** MODULES */
        public Communication getCommunicationInstance() {
        	if (communication == null)
				try {
					communication = new NetworkCommunication();
				} catch (IOException e) {
					throw new RuntimeException("Network initialization failed");
				}
        	return communication;
        }
        
        public SearchAnswerModul getSearchAnswerInstance() {
        	if (searchAnswerModule == null) 
        		searchAnswerModule = new SearchAnswerModul();
        	return searchAnswerModule;
        }
        
        public SearchModul getSearchInstance() {
        	if (searchModule == null)
        		searchModule = new SearchModul();
        	return searchModule;
        }
        
        public DownloadModul getDownloadInstance() {
        	if (downloadModul == null)
        		downloadModul = new DownloadModul();
        	return downloadModul;
        }
        
        public UploadModul getUploadInstance() {
        	if (uploadModul == null)
        		uploadModul = new UploadModul();
        	return uploadModul;
        }
        
        public PeerManagementModul getPeerManagementInstance() {
        	if (peerManagementModul == null)
        		peerManagementModul = new PeerManagementModul();
        	return peerManagementModul;
        }
        
    /** SELF IDENTIFICATION **/
        /* my peer id */
        public PeerID myPeerID() {
                return this.myPeerID;
        }
        
        /* my port */
        public int myPort() {
                return myPort;
        }
        
    /** PEERLIST MANIPULATION METHODS **/
                	
        /* add peer */
        /* NOTE! If a peer with the same ip + port but different id exists, this will replace him! */
        public void addPeer(PeerInfo peer) {
                peerList.addPeer(peer);
        }
        
        /* remove a peer */
        public synchronized boolean removePeer(PeerInfo peer) {
                PeerList.IPAddress addr = new PeerList.IPAddress(peer.ip(), peer.port());
            
                if (peerList.removePeerByAddress(addr)) {
                    rdList.addToList(addr);
                    return true;
                } else {
                    return false;
                }
        }
        
        /* remove peer by address */
        public boolean removePeerByAddress(InetAddress ip, int port) {
                return removePeer(new PeerInfo(null, ip, port));
        }
        
        /* clear peerlist */
        public synchronized void removeAllPeers() { synchronized (peerList) {
                PeerInfo[] p = getPeerList();
                
                for (int i = 0; i < p.length; i++)
                        rdList.addToList(new PeerList.IPAddress(p[i].ip(), p[i].port()));
                
                peerList.removeAll();
        } }
        
        /* get peerlist as an array */
        public PeerInfo[] getPeerList() {
                return peerList.getPeerList();
        }
        
        public int getPeerCount(){
        	// TODO ineffective;
        	return peerList.getPeerList().length;
        }
        
        /* get only active peers */
        public PeerInfo[] getActivePeers() {
                return peerList.getActivePeers();
        }
        
        public int getActivePeerCount() {
        	//TODO ineffective
        	return peerList.getActivePeers().length;
        }
        
        /* save peerlist to a specified file */
        public void savePeersToFile(String fileName) throws IOException {
                peerList.saveToFile(fileName);
        }
        
        /* load peerlist from a specified file */
        public void loadPeersFromFile(String fileName) throws IOException {
                peerList.loadFromFile(fileName, true);
        }
        
        /* save peerlist safely */
        public boolean backupPeerList() {
                try {
                        if (!lastPeerListBackup)
                                savePeersToFile(configPath() + peerFile1);
                        else
                                savePeersToFile(configPath() + peerFile2);
                        
                        lastPeerListBackup = !lastPeerListBackup;
                        return true;
                } catch (IOException e) {
                        return false;
                }
        }
        
        /* load peerlist from backup */
        public boolean loadPeerListBackup() {
                try {
                        peerList.loadFromFile(configPath() + peerFile1, true);
                        peerList.loadFromFile(configPath() + peerFile2, false);
                        
                        return true;
                } catch (IOException e) {
                        System.out.println("*** Error loading peerlist from file. ***");
                        return false;
                }
        }
        
        /* update PeerID in a certain PeerInfo */
        public boolean updatePeerID(PeerInfo peer, PeerID newID) {
                return peerList.updatePeerID(peer, newID);
        }
        
        /* get PeerInfo from peerlist if you know the PeerID */
        public PeerInfo findPeerByID(PeerID id) {
                return peerList.findPeerByID(id);
        }
    
    /** SEARCH RESULT MANIPULATION FUNCTIONS **/
            
        /* apply for receiving "new file found" messages */
        public void receiveSearchMessages(SearchResultReceiver rec) {
                resultList.addReceiver(rec);
        }
            
        /* add a single search result */
        public boolean addSearchResult(SearchResult srt) {
                return resultList.addResult(srt);
        }
        
        /* add multiple search results */
        public void addSearchResults(SearchResultList srl) {
                resultList.mergeWith(srl);
        }
        
        /* delete a single result */
        /*public boolean deleteSearchResult(SearchResult srt) {
                return resultList.removeResult(srt);
        }*/
        
        /* delete all results */
        public void deleteSearchResults() {
        	resultList.removeAll();
        }
        
        /* get all the search results sorted by filename & checksum */
        public SearchResult[] getSearchResults() {
                return resultList.getResults();
        }
        
        /* get all the search results sorted by filename & checksum */
        /* listing only one item for every group having the same    */
        /* filename and filesize and checksum                       */
        public SearchResult[] getUniqueSearchResults() {
                return resultList.getUniqueResults();
        }
        
        /* find all results with the given filename, listing only */
        /* unique files, i.e. one item per one size & checksum    */
        public SearchResult[] findResultsByFilename(String filename) {
                return resultList.findFileByName(filename);
        }
        
        /* find all peers having the given search result */
        /* (i.e. the same filename and checksum)         */
        public PeerID[] findOwnersOfFile(SearchResult str) {
                return resultList.findAllOwnersOfFile(str);
        }
        
        /* like getSearchReults(), but the results are given */
        /* in the order they appeared during the search      */
        public SearchResult[] getSearchResultsSortedByTime() {
                return resultList.getResultsSortedByTime();
        }
        
    /** STREAMID MANIPULATION METHODS **/
            
        /* generate new StreamID */
	public StreamID generateNewStreamID() {
                return new StreamID(streamIDGenerator.generateNewID());
	}
        
        /* return a used StreamID */
        public void returnUsedStreamID(StreamID id) {
                streamIDGenerator.returnUsedID(id.value());
        }
        
    /** MESSAGE COLLECTORS **/    
        public void raiseError(String description) {
        	errorCollector.raiseError(description);
        }
        
        public void addErrorCollectorObserver(ErrorCollectorObserver ob) {
        	errorCollector.addObserver(ob);
        }
        
        public void debugMessage(String description) {
        	debugCollector.throwMessage(description);
        }
        
        public void addDebugMessagesObserver(DebugCollectorObserver ob) {
        	debugCollector.addObserver(ob);
        }
        
    /** DOWNLOAD **/
        public DownloadedFilesDatabase DownloadedFiles(){
        	return this.down_files_db;
        }
        
        public SharedFilesDatabase SharedFiles(){
        	return this.shared_files_db;
        }
        
        public int activeDownloads() {
        	return getDownloadInstance().activeThreads();
        }
        
        public int activeUploads() {
        	return getUploadInstance().activeThread();
        }
        
        
    /** SINGLETONNESS - RELATED SAUCE **/
        
	protected static DataStorage _instance = null;
	public static DataStorage getInstance() {
		if (_instance == null) _instance = new DataStorage();
		return DataStorage._instance;
	}
	
	public static void setPort(int p) {
		if (portSet) return;
		port = p;
		portSet = true;
	}
    
    /** RANDOM NUMBER GENERATOR **/
        
        /* I'll put this here as Random class should be a singleton too */
        public byte getRandomByte() {
                return (byte)random.nextInt(256);
        }
    
    /** CONFIG FILES DIRECTORY **/
            
        public String configPath() {
                return configDir.toString() + "/";
        }
    
    /** RECENTLY DELETED PEERS LIST **/
        
        public void clearRecentlyDeleted() {
                rdList.clearList();
        }
        
        public boolean wasRecentlyDeleted(PeerInfo peer) {
                return rdList.isInList(new PeerList.IPAddress(peer.ip(), peer.port()));
        }
    /** DEBUG **/
        public void setDebug(boolean active) {
        	debugActive = active;
        }
        
        public boolean isDebugActive() {
        	return debugActive;
        }
        
    /** PROGRESS **/
        public void applyAsProgressSignalReceiver(ProgressSignalReceiver psr) {
        		currentProgressSignalReceiver = psr;
        }
            
        public void signalDownloadProgress() {
        	if (currentProgressSignalReceiver != null)
               currentProgressSignalReceiver.receiveProgressSignal(down_files_db.getTotalProgress());
        }

}
