package modules;

import auxiliary.Convert;
import auxiliary.SHA1;
import types.DownloadedFileDescriptor;
import types.Message;
import types.Message.MessageDataRepresentation;

/* Main data handlers of the download / upload communication. 
 * 
 * */

/* Sub-Protocol description 
 * ==Participating classes:
 * --DownloadModul (receives messages from Communication and passes them to DownloadThread)
 * --DownloadThread (responsible for all manipulation with downloaded file)
 * --Upload{the same} (the same on the other side)
 * --DownloadedFileDescriptor (remembers all what is needed about the downloaded resource on the downloader side)
 * --SharedFileDescriptor (on the uploaders side -- same as DownloadedFileDescriptor) 
 * --Download/SharedFiles{Database} -- stores Descriptors
 * --DownloadMessage -- parses / encodes messages sent between Upload/Download{Module}
 * 
 * Optimal progress:
 * D->U Init(SequenceID, Checksum)
 * U->D Init(FileSize)
 * D->U Part(part_id) --repeat until all parts  (part_data = null)
 * U->D Part(part_id, part_data)
 * 
 * */

public class DownloadMessage extends Message {
	public enum Type{
		INIT (0x04),
		PART (0x05),
		CANCEL (0x06);
		private final int code;  
		Type(int code){
			this.code = code;
		}
		byte getCode() {
			return (byte)this.code;
		}
	}
	public enum InitType{
		OK (0x00),
		REQUEST (0x01),
		REGEN (0x02),
		CANCEL (0x03);
		private final int code;
		InitType(int code){
			this.code=code;
		}
		byte getCode(){
			return (byte)this.code;
		}
	}
	public enum CancelType{
		FINISHED (0x00),
		NOT_SPECIFIED (0x01),
		FILE_NOT_EXISTS (0x02);
		private final int code;
		CancelType(int code){
			this.code=code;
		}
		byte getCode(){
			return (byte)this.code;
		}
	}

	public static DownloadMessage.CancelType getCancelTypeFromByte(byte b){
		switch(b){
		case 0x00: return CancelType.FINISHED;			//Downloader sends it. 
		case 0x01: return CancelType.NOT_SPECIFIED;		//Both sides (like user cancel, something was not successful (file save, ...)) 
		case 0x02: return CancelType.FILE_NOT_EXISTS;	//Uploader sends it. 
		}
		return CancelType.NOT_SPECIFIED;
	}
	
	//Message.data -> class data
	public static void parse(Message message) {
		DownloadMessage factory = new DownloadMessage();
		
		if(message.messageType == DownloadMessage.Type.INIT.getCode()){
			message.representation = factory.new InitMessage();
		}else if(message.messageType == DownloadMessage.Type.PART.getCode()){
			message.representation = factory.new PartMessage();
		}else if(message.messageType == DownloadMessage.Type.CANCEL.getCode()){
			message.representation = factory.new CancelMessage();
		}else {
			DataStorage.getInstance().debugMessage("DownloadMessage.parse haven't found MessageDataRepresentation for MessageType=" + message.messageType);
			message.representation = null;
			return;
		}
		
		message.representation.parse(message.data);
	}
	//TODO
	/*
	public Message encodeAndReply(){
		return 
	}*/
	
	public class InitMessage implements MessageDataRepresentation{
		
		public InitType init_type; 
		public SHA1 init_sha = null; 
		public long file_size; //in response(ok)
		@Override
		public boolean parse(byte[] data) {
			if(data[0] == InitType.OK.getCode()) {
				this.init_type = InitType.OK;
				this.file_size = Convert.bytesToLong(data, 1);
			}
			if(data[0] == InitType.REQUEST.getCode()) {
				this.init_type = InitType.REQUEST;
				
				byte[] sha_data = new byte[SHA1.SHA1_LENGTH];
				System.arraycopy(data, 1, sha_data, 0, SHA1.SHA1_LENGTH);
				this.init_sha = new SHA1(sha_data);
				
				//System.out.println("Request parsed sha: " + this.init_sha.toString());
			}
			if(data[0] == InitType.REGEN.getCode()) this.init_type = InitType.REGEN;
			if(data[0] == InitType.CANCEL.getCode()) this.init_type = InitType.CANCEL;
			return true;
		}
		@Override
		public byte[] encode() {
			byte[] result = null;
			if(this.init_type == InitType.OK){
				result = new byte[9];
				System.arraycopy(Convert.longToBytes(file_size), 0, result, 1, 8);
			}
			if(this.init_type == InitType.REQUEST){
				result = new byte[SHA1.SHA1_LENGTH + 1];
				System.arraycopy(init_sha.value(), 0, result, 1, SHA1.SHA1_LENGTH);
			}
			if(this.init_type == InitType.REGEN){
				result = new byte[1];
			}
			if(this.init_type == InitType.CANCEL){
				result = new byte[1];
			}
			result[0] = this.init_type.getCode();
			return result;
		}
		@Override
		public byte getMessageType() {
			return DownloadMessage.Type.INIT.getCode();
		}
	}
	
	public class PartMessage implements MessageDataRepresentation{
		//part_id < 0 --> treat as null
		public int part_id;
		public byte[] part_data = null;
		
		@Override
		public boolean parse(byte[] data) {
			//Request
			if(data.length == 4){
				this.part_id = Convert.bytesToInt(data, 0);
				this.part_data = null;
			}
			//Response
			else{
				this.part_id = Convert.bytesToInt(data, 0);
				this.part_data = new byte[DownloadedFileDescriptor.part_size];
				System.arraycopy(data, 4, this.part_data, 0, DownloadedFileDescriptor.part_size);
			}
			return true;
		}
		@Override
		public byte[] encode() {
			byte[] result = null;
			//Request
			if(this.part_data == null){
				result = Convert.intToBytes(this.part_id);
			}
			//Response
			else{
				result = new byte[4 + DownloadedFileDescriptor.part_size];
				System.arraycopy(Convert.intToBytes(this.part_id), 0, result, 0, 4);
				System.arraycopy(this.part_data, 0, result, 4, DownloadedFileDescriptor.part_size);
			}
				
			return result; 
		}
		@Override
		public byte getMessageType() {
			return DownloadMessage.Type.PART.getCode();
		}
	}
	
	public class CancelMessage implements MessageDataRepresentation{
		public CancelType cancel_type;
		
		@Override
		public boolean parse(byte[] data) {
			if(data == null){
				DataStorage.getInstance().debugMessage("CancelMessage data is null.");
				this.cancel_type = CancelType.NOT_SPECIFIED;
			}
			else{
				this.cancel_type = DownloadMessage.getCancelTypeFromByte(data[0]);
			}
			return true;
		}

		@Override
		public byte[] encode() {
			byte[] result = new byte[1];
			result[0] = this.cancel_type.getCode();
			return result;
		}
		@Override
		public byte getMessageType() {
			return DownloadMessage.Type.CANCEL.getCode();
		}

		public String cancelReason() {
			switch(this.cancel_type.getCode()){
			case 0x00:  return "All parts have been downloaded.";
			case 0x02:  return "Requested file not exists (on the uploaders side).";
			case 0x01:  
			default:	return "All parts have been downloaded.";
			}
		}
	}
}
