package modules;

import java.io.IOException;
import java.io.OutputStream;
import java.net.*;
import java.util.HashMap;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;

import types.*;

// TODO better stream parsing in TCPClientListener
// TODO always set remote peer in message passed to modules
// TODO exceptions when tcp connecntion is closed

public class NetworkCommunication implements Communication {

	private int MESSAGE_BUFFER_SIZE = 100;		
	
	private ServerSocket tcpServerSocket;
	private HashMap<StreamID, TCPClientListener> tcpClientListeners;
	private UDPListener udpListener;
	
	private boolean exit = false;
	HashMap<ModulType, ArrayBlockingQueue<Message> > storage;
	private DataStorage dataStorage;
	
//------------------ CONSTRUCTORS & DESTRUCTORS ------------------//
	
	public NetworkCommunication() throws IOException {
		this.dataStorage = DataStorage.getInstance();
		
		storage = new HashMap<ModulType, ArrayBlockingQueue<Message>>();
		for (ModulType m: ModulType.values())
			storage.put(m, new ArrayBlockingQueue<Message>(MESSAGE_BUFFER_SIZE, true));
		
		udpListener = new UDPListener(dataStorage, storage, dataStorage.myPort());
		tcpServerSocket = new ServerSocket(dataStorage.myPort());
		tcpClientListeners = new HashMap<StreamID, TCPClientListener>();
	}

//---------------------- MODULE METHODS --------------------------//
	public void removeListener(StreamID sid) {
		tcpClientListeners.remove(sid);
	}
	
	@Override
	public void run() {
		Thread newUdpListener = new Thread(udpListener);
		newUdpListener.start();
		
		while (!exit) {			
			try {
				Socket tcpSocket = tcpServerSocket.accept();
				StreamID streamId = dataStorage.generateNewStreamID();
				
				//dataStorage.debugMessage("recieved TCP connection from " + tcpSocket.getInetAddress().getHostAddress());
				//dataStorage.debugMessage("socket " + (tcpSocket.isBound() ? "is" : "is not") + " bound");
				//dataStorage.debugMessage("socket " + (tcpSocket.isClosed() ? "is" : "is not") + " closed");
				
				TCPClientListener listener = new TCPClientListener(dataStorage, storage, this, tcpSocket, streamId);
				Thread newListener = new Thread(listener);
				newListener.start();
					
				tcpClientListeners.put(streamId, listener);
			} catch (IOException e) {
				continue;
			}
		}
		
		try {
			tcpServerSocket.close();
		} catch (IOException e) {}
		
		udpListener.exit();
		
		for (TCPClientListener listener: tcpClientListeners.values())
			listener.exit();
	}

	@Override
	public void exit() {
		exit = true;
		try {
			tcpServerSocket.close();
		} catch (IOException e) {}
	}
	
	
	
//-------------------- COMMUNICATION METHDOS ---------------------//
	@Override
	public Message receiveMessage(ModulType modul) throws InterruptedException {
		Message msg = storage.get(modul).take();
		if (msg.streamID != null){
			//--The greatest part of the app code :)
			//dataStorage.debugMessage(modul.toString() + " picked message " + ((msg == null) ? "null" : ""));
			//dataStorage.debugMessage("Received message:\n"+msg.toString());
		}
		return msg;
	}
	
	@Override
	public Message receiveMessage(ModulType modul, long timeout, TimeUnit unit) throws InterruptedException {
		Message msg = storage.get(modul).take();
		if (msg.streamID != null){
			//dataStorage.debugMessage(modul.toString() + " picked message " + ((msg == null) ? "null" : ""));
			//dataStorage.debugMessage("Received message:\n"+msg.toString());
		}
		return msg;
	}

	@Override
	public void sendMessage(Message msg) throws IOException {
		byte[] bytes = msg.getBytes();
		if (msg.streamID == null) {
			DatagramPacket udpPacket = new DatagramPacket(bytes, 0, bytes.length);
			udpPacket.setAddress(msg.remotePeer.ip());
			udpPacket.setPort(msg.remotePeer.port());
			udpListener.udpSocket().send(udpPacket);
		}
		else {
			if (!tcpClientListeners.containsKey(msg.streamID)) {
				if (msg.streamClosed) {
					dataStorage.raiseError("Trying to close non-existing stream");
					return;
				}
				
				Socket tcpSocket = new Socket(msg.remotePeer.ip(), msg.remotePeer.port());
				
				TCPClientListener listener = new TCPClientListener(dataStorage, storage, this, tcpSocket, msg.streamID, msg.targetModule());
				Thread newListener = new Thread(listener);
				newListener.start();
				
				tcpClientListeners.put(msg.streamID, listener);
			}
			
			Socket tcpSocket = tcpClientListeners.get(msg.streamID).tcpSocket();
			if (tcpSocket.isClosed()) {
				dataStorage.raiseError("WARNING: trying to use closed socket.");
				TCPClientListener lst = tcpClientListeners.get(msg.streamID);
				if(lst != null) lst.exit();
				tcpClientListeners.remove(msg.streamID);
				return;
			}
			
			if (msg.streamClosed) {
				tcpSocket.shutdownOutput();
			}
			else {
				OutputStream out = tcpSocket.getOutputStream();
				/*
		        at java.net.SocketOutputStream.socketWrite0(Native Method)
		        at java.net.SocketOutputStream.socketWrite(SocketOutputStream.java:92)
		        at java.net.SocketOutputStream.write(SocketOutputStream.java:124)
		        at modules.NetworkCommunication.sendMessage(NetworkCommunication.java:146 -- next line)
		        at modules.UploadModul.sendMessage(UploadModul.java:60)
		        at modules.UploadThread.run(UploadThread.java:109)
		        at java.lang.Thread.run(Thread.java:662)
				*/
				out.write(bytes);
				if (msg.streamID != null){
					//dataStorage.debugMessage("Sent message: \n" + msg.toString());
				}
			}
		}
	}
}
