package modules;

import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Collections;
import java.util.Iterator;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Scanner;
import java.io.BufferedReader;
import java.io.FileReader;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.TreeMap;
import java.util.SortedMap;
import java.lang.Comparable;

import types.*;

/** PeerList class            **/
/** used in DataStorage class **/

public class PeerList {

    // address + port pair
    static class IPAddress implements Comparable<Object> {
            public InetAddress addr;
            public int port;
        
            IPAddress(InetAddress addr, int port) {
                this.addr = addr;
                this.port = port;
            }
        
            @Override
            public boolean equals(Object obj) {
                    if (! (obj instanceof IPAddress) )
                            throw new ClassCastException();

                    return (compareTo(obj) == 0);
            }
            
            @Override
            public int compareTo(Object obj) {
                    if (! (obj instanceof IPAddress) )
                            throw new ClassCastException();

                    if (! this.addr.equals(((IPAddress) obj).addr)) {
                            byte[] myBytes = this.addr.getAddress();
                            byte[] theirBytes = ((IPAddress) obj).addr.getAddress();

                            for (int i = 0; i < 4; i++)
                                    if (myBytes[i] != theirBytes[i])
                                            return (int)myBytes[i] - (int)theirBytes[i];
                    }
                
                    if (this.port < ((IPAddress) obj).port) return -1;
                    if (this.port > ((IPAddress) obj).port) return 1;
                    
                    return 0;
            }
    }
    
    // internal representation
    SortedSet<PeerInfo> list;
    SortedMap<IPAddress, PeerInfo> addrLookup;
    
    // constructor
    public PeerList() {
            list = Collections.synchronizedSortedSet(new TreeSet<PeerInfo>());
            addrLookup = Collections.synchronizedSortedMap(new TreeMap<IPAddress, PeerInfo>());
    }
    
    private static InetAddress minInetAddress() throws UnknownHostException {
            byte[] addr = new byte[4];
            
            for (int i = 0; i < 4; i++)
                    addr[i] = Byte.MIN_VALUE;
            
            return InetAddress.getByAddress(addr);
    }
    
    // add peer to the list
    public synchronized void addPeer(PeerInfo peer) {
            if (peer == null)
                    throw new NullPointerException();
        
            IPAddress addr = new IPAddress(peer.ip(), peer.port());
            if (addrLookup.containsKey(addr))
                    removePeerByAddress(addr);

            list.add(peer);
            addrLookup.put(addr, peer);
    }
    
    // remove a peer from the list
    public synchronized boolean removePeerByAddress(IPAddress addr) {
            if (!addrLookup.containsKey(addr))
                    return false;
            
            list.remove(addrLookup.get(addr));
            addrLookup.remove(addr);
            
            return true;
    }
    
    // remove all peers
    public synchronized void removeAll() {
            list.clear();
            addrLookup.clear();
    }
        
    // get all peers - this will be reimplemented after
    // we decide on the bucket / whatever system
    public synchronized PeerInfo[] getPeerList() {
            PeerInfo[] ret = new PeerInfo[list.size()];
            Iterator<PeerInfo> iter = list.iterator();
                
            for (int i = 0; i < list.size(); i++)
                    ret[i] = iter.next();
                
            return ret;
    }
    
    // get only the active peers
    public synchronized PeerInfo[] getActivePeers() {
            try {
                    SortedSet<PeerInfo> actives = list.tailSet(new PeerInfo(new PeerID(Integer.MIN_VALUE),
                                                               minInetAddress(), Integer.MIN_VALUE));
                    Iterator<PeerInfo> iter = actives.iterator();
                    PeerInfo[] ret = new PeerInfo[actives.size()];
            
                    for (int i = 0; i < actives.size(); i++)
                            ret[i] = iter.next();

                    return ret;
            } catch (UnknownHostException e) {
                    return new PeerInfo[0];
            }
    }
    
    // load peerlist from file
    public synchronized void loadFromFile(String fileName, boolean clearFirst) throws IOException {
            Scanner in = new Scanner(new BufferedReader(new FileReader(fileName)));
            if (clearFirst) list.clear();
            
            while (in.hasNext()) {
                InetAddress addr = InetAddress.getByName(in.next());
                int port = in.nextInt();
                addPeer(new PeerInfo((PeerID)null, addr, port));
            }
            
            in.close();
    }
    
    // save peerlist to file
    public synchronized void saveToFile(String fileName) throws IOException {
            PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(fileName)));
            Iterator<PeerInfo> iter = list.iterator();
            PeerInfo current;
            
            while (iter.hasNext()) {
                    current = iter.next();
                    out.println(current.ip().getHostAddress() + " " + new Integer(current.port()).toString());
            }
            
            out.close();
    }
    
    // update peer ID
    public synchronized boolean updatePeerID(PeerInfo peer, PeerID newID) {
            PeerInfo newPeer = new PeerInfo(newID, peer.ip(), peer.port());
            
            boolean ret = addrLookup.containsKey(new IPAddress(peer.ip(), peer.port()));
            
            addPeer(newPeer);
            
            return ret;
    }
    
    // find entry by ID
    public synchronized PeerInfo findPeerByID(PeerID id) {
            try {
                    SortedSet<PeerInfo> span = list.tailSet(new PeerInfo(id, minInetAddress(),
                                                                         Integer.MIN_VALUE));
                    if (span.iterator().hasNext() && (span.iterator().next().id().equals(id)))
                            return span.iterator().next();
                    else
                            return null;

            } catch (Exception e) {
                    DataStorage.getInstance().raiseError("*** Couldn't find peer by ID; " +
                                                           "A null value was supplied instead of the id. ***");
                    return null;
            }
    }
    
}
