package modules;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import types.Message;
import types.ModulType;
import types.PeerInfo;
import types.SearchID;
import types.SearchQuery;
import types.SearchResult;
import types.SequenceID;
import types.SharedFileDescriptor;

/*
 * processes search queries, from network
 * 
 * SM listens to NC. When query comes, SAM saves its ID, responds with results to sender, and forwardes this query to neigbours
 * When comes query twice, ignores it
 */

public class SearchAnswerModul implements Module {
	
	private static long BLOCKING_TIME = 3600000; // in miliseconds
	
	private DataStorage ds;
	private Communication comm;
	
	private boolean exit = false;
	private Thread thread;
	
	private HashMap<SearchID, Long> searches;
	
	public SearchAnswerModul() {
		this.ds = DataStorage.getInstance();
		this.comm = ds.getCommunicationInstance();
		searches = new HashMap<SearchID, Long>();
	}
	
	@Override
	public void run() {
		thread = Thread.currentThread();
		
		while (!exit) {
			try {
				Message msg = comm.receiveMessage(ModulType.SEARCH_ANSWER);
				processMessage(msg);
			} catch (InterruptedException e) {
				continue;
			}
		}
	}
	
	@Override
	public void exit() {
		exit = true;
		if (thread != null) thread.interrupt();
	}
	
	//
	
	private void processMessage(Message msg) {
		SearchID sid = new SearchID(msg.identifier.value());
		if (isQueryRedundant(sid)) return;
		
		SearchQuery sq = SearchQuery.querryWithBytes(msg.data);
		if (sq == null) return;
		
		ds.debugMessage("recieved query (ttl = "+ sq.ttl +") from " + msg.remotePeer.ip().getHostAddress());
		
		rememberQuery(sid);
		answerQuery(sq, msg);
		if (sq.ttl > 0) forwardQuery(sq, sid, msg.remotePeer);
	}
	
	private void answerQuery(SearchQuery sq, Message msg) {
		SharedFileDescriptor[] files = ds.SharedFiles().filesByKeyWords(sq.keywords);
		for (SharedFileDescriptor sfd: files) {
			
			SearchResult sr = SearchResult.fromSharedFileDescriptor(sfd);
			if (sr == null) continue;
			
			Message nmsg =  new Message();
			nmsg.identifier = msg.identifier;
			nmsg.messageType = 0x03;
			if (sq.initiator.port() == 0) nmsg.remotePeer = msg.remotePeer;
			else nmsg.remotePeer = sq.initiator;
			nmsg.sourceID = ds.myPeerID();
			nmsg.targetID = nmsg.remotePeer.id();
			nmsg.streamClosed = false;
			nmsg.streamID = null;
			nmsg.data = sr.getBytes();
			if (nmsg.data == null) continue;
			nmsg.dataLength = nmsg.data.length;
			
			try {
				ds.debugMessage("sent answer to "+nmsg.remotePeer.ip().getHostAddress());
				comm.sendMessage(nmsg);
			} catch (IOException e) {
				continue;
			}
		}
	}
	
	private void forwardQuery(SearchQuery sq, SearchID sid, PeerInfo from) {
		
		sq.ttl--;
		if (sq.initiator.port() == 0) sq.initiator = from;
		
		Message msg =  new Message();
		msg.identifier = new SequenceID(sid.value());
		msg.messageType = 0x02;
		msg.sourceID = ds.myPeerID();
		msg.streamClosed = false;
		msg.streamID = null;
		msg.data = sq.getBytes();
		msg.dataLength = msg.data.length;
		
		int sent = 0;
		for (PeerInfo pi: ds.getActivePeers()) {
			//if (sent >= SearchModul.BRANCH_FACTOR) break
			if (pi.id().equals(from.id())) continue;					// dont forward back to sender
			
			Message m = msg.deepCopy();
			m.remotePeer = pi;
			m.targetID = pi.id();
			try {
				ds.debugMessage("query (ttl = "+ sq.ttl +") forwarded to " + pi.ip().getHostAddress());
				comm.sendMessage(m);
			} catch (IOException e) {
				continue;
			}
			sent++;
		}
	}
	
	private boolean isQueryRedundant(SearchID sid) {
		clearOldSearches();
		if (!searches.containsKey(sid)) return false;	
		return true;
	}
	
	private void rememberQuery(SearchID sid) {
		searches.put(sid, System.currentTimeMillis());
	}
	
	private void clearOldSearches() {
		long currentTime = System.currentTimeMillis();
		for (Iterator<Map.Entry<SearchID, Long>> it = searches.entrySet().iterator(); it.hasNext(); ) {
			long time = it.next().getValue();
			if (time + BLOCKING_TIME < currentTime)
				it.remove();
		}
	}
}
