package modules;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Random;

import types.*;

/*
 * processes search queries from user
 * 
 * When user initiates search, SM sends query to NC, and listens to answers from NC (parallel searches allowed)
 * when answers comes, SM processes it, and add results to DS.
 * when search is killed, SM starts ignoring its messages.
 */

// SINGLETON

public class SearchModul implements Module {

	public static int BRANCH_FACTOR = 5;
	public static int SEARCH_DEPTH = 5;
	
	private DataStorage ds;
	private Communication comm;
	private boolean exit = false;
	private Thread thread;
	
	private SearchID actualSearch;
	
	public SearchModul() {
		this.ds = DataStorage.getInstance();
		this.comm = ds.getCommunicationInstance();
	}
	
	// query does not have to have filled ttl nor initiator...
	public SearchID search(SearchQuery sq) {
		if (sq == null) throw new NullPointerException();
		if (sq.keywords == null) throw new IllegalArgumentException();
		
		sq.ttl = SEARCH_DEPTH;
		try {
			sq.initiator = new PeerInfo(new PeerID(0),InetAddress.getByName("0.0.0.0"), 0);
		} catch (UnknownHostException e1) {
			return null;
		}
		
		Random rand = new Random();
		actualSearch = new SearchID(rand.nextInt());
		ds.deleteSearchResults();
		
		Message msg =  new Message();
		msg.identifier = new SequenceID(actualSearch.value());
		msg.messageType = 0x02;
		msg.sourceID = ds.myPeerID();
		msg.streamClosed = false;
		msg.streamID = null;
		msg.data = sq.getBytes();
		msg.dataLength = msg.data.length;
		
		int sent = 0;
		for (PeerInfo pi: ds.getActivePeers()) {
			//if (sent >= BRANCH_FACTOR) break;
			Message m = msg.deepCopy();
			m.remotePeer = pi;
			m.targetID = pi.id();
			try {
				comm.sendMessage(m);
				ds.debugMessage("sent query (ttl = "+ sq.ttl +") to: "+pi.ip().getHostAddress());
			} catch (IOException e) {
				continue;
			}
			sent++;
		}
		
		
		return actualSearch;
	}
	
	public void killSearch(SearchID sid) {
		// TODO if we start supporting multiple searches
	}
	
	@Override
	public void run() {
		thread = Thread.currentThread();
		
		while (!exit) {
			try {
				Message msg = comm.receiveMessage(ModulType.SEARCH);
				processMessage(msg);
				
			} catch (InterruptedException e) {
				continue;
			}	
		}
	}
	
	@Override
	public void exit() {
		exit = true;
		if (thread != null) thread.interrupt();
	}
	
	//------- AUXILIARY
	
	private void processMessage(Message msg) {
		SearchID sid = new SearchID(msg.identifier.value());
		SearchResult sr = SearchResult.fromMessage(msg);
		if (sr == null) return;
		
		if (!sid.equals(actualSearch)) return;
		
		ds.debugMessage("recieved result from: " + msg.remotePeer.ip().getHostAddress());
		
		ds.addSearchResult(sr);
	}

}
