package modules;

import java.util.TreeSet;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.ArrayList;
import java.util.Collections;
import java.util.TreeMap;
import java.util.SortedMap;

import auxiliary.SHA1;

import types.*;

public class SearchResultList {
    
    /** GENERAL MEMBERS **/
            
        /* private members */
        private SortedSet<SearchResult> list;
        private SortedMap<Integer, SearchResult> timeSorted;
        private ArrayList<SearchResultReceiver> receivers;
        
        /* constructor */
        public SearchResultList() {
                list = Collections.synchronizedSortedSet(new TreeSet<SearchResult>());
                timeSorted = Collections.synchronizedSortedMap(new TreeMap<Integer, SearchResult>());
                receivers = new ArrayList<SearchResultReceiver>();
        }
        
    /** MESSAGES **/
            
        /* add a "new file found" messages receiver */
        public synchronized void addReceiver(SearchResultReceiver rec) {
            receivers.add(rec);
        }
        
        /* send the message to the receivers */
        private synchronized void sendMessageToReceivers() {
            Iterator<SearchResultReceiver> iter = receivers.iterator();
            
            while (iter.hasNext())
                iter.next().receiveSearchMessage();
        }
        
    /** ADDING, REMOVING... **/
        
        /* add a single item */
        public synchronized boolean addResult(SearchResult srt) {
                timeSorted.put(list.size(), srt);
                boolean ret = list.add(srt);
                sendMessageToReceivers();
                return ret;
        }
        
        /* remove a single item */
        /*public synchronized boolean removeResult(SearchResult srt) {
                timeSorted.remove(srt);
                return list.remove(srt);
        }*/
        
        /* remove all items */
        public synchronized void removeAll() {
                timeSorted.clear();
        	list.clear();
                sendMessageToReceivers();
        }
        
        /* merge lists */
        /* A.mergeWith(B) means A = A u B; B remains unchanged */
        public synchronized void mergeWith(SearchResultList srt) {
                for (SearchResult s : srt.getResults())
                        addResult(s);
                sendMessageToReceivers();
        }
        
    /** LISTING **/
            
        /* get list as an array */
        public synchronized SearchResult[] getResults() {
            Iterator<SearchResult> iter = list.iterator();
            
            SearchResult[] ret = new SearchResult[list.size()];
            
            for (int i = 0; i < list.size(); i++)
                ret[i] = iter.next();
            
            return ret;
        }
        
        /* get list as an array, but with duplicates merged */
        public synchronized SearchResult[] getUniqueResults() {
                Iterator<SearchResult> iter = list.iterator();
            
                if (list.size() == 0)
                        return new SearchResult[0];
            
                int itemCount = 1;
                SearchResult lastItem = iter.next();
                SearchResult currentItem;
            
                for (int i = 1; i < list.size(); i++)
                        if (!(currentItem = iter.next()).equalFile(lastItem)) {
                            itemCount++;
                            lastItem = currentItem;
                        }
            
                SearchResult[] ret = new SearchResult[itemCount];
            
                itemCount = 1;
                iter = list.iterator();
                ret[0] = lastItem = iter.next();
            
                for (int i = 1; i < list.size(); i++)
                        if (!(currentItem = iter.next()).equalFile(lastItem)) {
                                ret[itemCount++] = currentItem;
                                lastItem = currentItem;
                        }
            
                return ret;
        }
        
        /* search by name (SearchResult defines a lexicographical + checksum order) */
        public synchronized SearchResult[] findFileByName(String filename) {
                SortedSet<SearchResult> span =  list.tailSet(new SearchResult(new PeerID(0), filename, 0, new SHA1()));
                Iterator<SearchResult> iter = span.iterator();
            
                int filesCount = 0;
                SearchResult current, previous = span.iterator().next();
            
                for (int i = 0; i < span.size(); i++) {
                        current = iter.next();
                        if (!current.filename().equals(filename)) break;
                        if (filesCount > 0 && current.equalFile(previous)) continue;
                        ++filesCount;
                        previous = current;
                }
                
                SearchResult[] ret = new SearchResult[filesCount];
                iter = span.iterator();
                
                for (int i = 0; i < filesCount; i++) {
                        current = iter.next();
                        if (i > 0 && current.equalFile(ret[i - 1])) {
                                --i;
                                continue;
                        }
                        ret[i] = current;
                }
                
                return ret;
        }
        
        /* find all owners of a particular file */
        public synchronized PeerID[] findAllOwnersOfFile(SearchResult srt) {
                SortedSet<SearchResult> span = list.tailSet(srt);
                Iterator<SearchResult> iter = span.iterator();
                
                SearchResult current;
                int ownersCount = 0;
                
                for (int i = 0; i < span.size(); i++) {
                        current = iter.next();
                        if (!current.equalFile(srt)) break;
                        ++ownersCount;
                }
                
                PeerID[] ret = new PeerID[ownersCount];
                iter = span.iterator();
                
                for (int i = 0; i < ownersCount; i++)
                        ret[i] = iter.next().peer();
                
                return ret;
        }
        
        /* results sorted by the time they were added */
        public synchronized SearchResult[] getResultsSortedByTime() {
                SearchResult[] ret = new SearchResult[list.size()];
            
                for (int i = 0; i < list.size(); i++)
                        ret[i] = timeSorted.get(new Integer(i));
                            
                return ret;
        }
}
