package modules;

//!TODO check if remove(Object o) --> mistype

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Scanner;

import auxiliary.Convert;
import auxiliary.SHA1;

import types.SharedFileDescriptor;

//Checksum => SharedFileDescriptor
//--Note: redundancy of checksum
//!TODO Resolve concurrency
public class SharedFilesDatabase{

	private static final long serialVersionUID = -6352569767246599722L;
	//Must be distinct from 0-9,a-f,filepath characters
	private static final String ini_file_delimiter = "\n";
	
	HashSet<SharedFileDescriptor> data = new HashSet<SharedFileDescriptor>();
	protected HashMap<SHA1, SharedFileDescriptor> indexByChecksum = new HashMap<SHA1, SharedFileDescriptor>(); 

	//Note that DownloadedFilesDatabase has other conditions to add a file into it. 
	public boolean add(SharedFileDescriptor sfd){
		//Non existing files cannot be added.
		if(!(new File(sfd.getFilepath())).exists())
			return false;
		
		if(this.containsChecksum(sfd.getTargetChecksum())){
			DataStorage.getInstance().debugMessage("Checksum " + sfd.getTargetChecksum() + " of file " + sfd.getFilepath() + " is already contained in the database.");
			return false;
		}
		
		if(data.add(sfd)){
			this.indexByChecksum.put(sfd.getTargetChecksum(), sfd);
			return true;
		}
		return false;
	}
	public boolean remove(SharedFileDescriptor sfd) {
		if(data.remove(sfd)){
			this.indexByChecksum.remove(sfd.getTargetChecksum());
			return true;
		}
		return false;
	}
	public int size(){
		return this.data.size();
	}
	
	public boolean contains(SharedFileDescriptor sfd){
		return this.data.contains(sfd);
	}
	
	public boolean containsChecksum(SHA1 checksum){
		return this.indexByChecksum.containsKey(checksum);
	}
	//!TODO -- to much coupling
	public HashSet<SharedFileDescriptor> getFiles(){
		return this.data;
	}
	
	//Called from ini file
	protected boolean addDescriptor(String filename, String folder, SHA1 hash){
		return this.add(new SharedFileDescriptor(filename, folder, hash));
	}
	
	public String configPath(){
		return DataStorage.getInstance().configPath() + DataStorage.shared_db_ini_file;
	}
	
	protected boolean loadFromFile(String filepath){
		if(!(new File(filepath).exists())){
			try {
				(new File(filepath)).createNewFile();
			} catch (IOException e1) {
				DataStorage.getInstance().debugMessage("Couldn't load shared.ini file: " + filepath);
			}
			//In any case nothing to read. 
			return false;
		}
		synchronized(this) {
			try {
				Scanner in = new Scanner(new BufferedReader(new FileReader(filepath)));
				in.useDelimiter(ini_file_delimiter);
				int count = in.nextInt();
				for(int i=0; i<count; i++){
					this.addDescriptor(in.next(), in.next(), new SHA1(Convert.alphaHexStringToBytes(in.next())));
				}
			} catch (FileNotFoundException e) {
				DataStorage.getInstance().debugMessage("Couldn't find shared.ini file: " + filepath);
				return false;
			} catch (Exception e){
				DataStorage.getInstance().debugMessage("Corrupted shared.ini file: " + filepath);
				return false;
			}
		}
		DataStorage.getInstance().debugMessage("ini.file (should be) successfully loaded: " + filepath);
		
		return true;
	}
	
	//Load from ini file
	public SharedFilesDatabase(){
		super();
	}
	//Will be overriden in DownloadDatabase -- it has its own init file.  
	//--must be called externally. 
	public void init(){
		this.loadFromFile(this.configPath());
	}
	
	public String getFilenameByChecksum(SHA1 checksum){
		if(!this.indexByChecksum.containsKey(checksum))
			return null;
		return this.indexByChecksum.get(checksum).getFilename();
	}
	public SharedFileDescriptor getDescriptorByChecksum(SHA1 checksum){
		if(!this.indexByChecksum.containsKey(checksum))
			return null;
		return this.indexByChecksum.get(checksum);
	}
	
	public SharedFileDescriptor[] filesByKeyWords(Collection<String> keywords) {
		
		HashSet<SharedFileDescriptor> files = new HashSet<SharedFileDescriptor>();
		for (SharedFileDescriptor sfd: this.data) {
			String filename = sfd.getFilename();
			for (String kw: keywords)
				if (filename.contains(kw)) files.add(sfd);
		}
		
		SharedFileDescriptor[] result = new SharedFileDescriptor[files.size()];
		int i = 0;
		for (SharedFileDescriptor sfd: files)
			result[i++] = sfd;
			
		return result;
	}

	public SharedFileDescriptor[] filesByFilename(String filename) {		
		if (filename == null) return new SharedFileDescriptor[0];
		
		ArrayList<SharedFileDescriptor> files = new ArrayList<SharedFileDescriptor>();
		for (SharedFileDescriptor sfd: this.data) {
			if (filename.equals(sfd.getFilename())) files.add(sfd);
		}
		
		SharedFileDescriptor[] result = new SharedFileDescriptor[files.size()];
		for (int i=0; i<files.size(); i++)
			result[i] = files.get(i);
			
		return result;
	}
	
	public SharedFileDescriptor fileByFilePath(String filepath) {
		if (filepath == null) return null;
		
		for (SharedFileDescriptor sfd: this.data)
			if (filepath.equals(sfd.getFilepath())) return sfd;
		
		return null;
	}
	
	protected boolean backupToFile(String filepath){
		PrintWriter out;
		try {
			out = new PrintWriter(new BufferedWriter(new FileWriter(filepath)));
		} catch (IOException e) {
			DataStorage.getInstance().debugMessage("Ini file " + this.configPath() + " NOT saved.");
			return false;
		}

		synchronized(this) {
			out.write(((Integer)this.data.size()).toString());
			out.write(ini_file_delimiter);
			for(SharedFileDescriptor d : this.data){
				out.write(d.getFilename());
				out.write(ini_file_delimiter);
				out.write(d.getFolder());
				out.write(ini_file_delimiter);
				out.write(d.getTargetChecksum().toString());
				out.write(ini_file_delimiter);
			}
			out.close();
		}
		DataStorage.getInstance().debugMessage("Ini file " + this.configPath() + " saved.");
		
		return true;
	}
	
	//Save to ini file
	//Can be called at runtime
	public boolean backup(){
		return this.backupToFile(this.configPath());
	}
	
	protected void finalize(){
		this.backup();
	}
}
