package modules;

import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;
import java.util.HashMap;
import java.util.concurrent.ArrayBlockingQueue;

import types.Message;
import types.ModulType;
import types.PeerInfo;
import types.StreamID;

public class TCPClientListener extends NetworkListener {
	
	private int TIMEOUT = 1000;
	private Thread thread = null;
	
	private Socket tcpSocket;
	private StreamID streamID;
	private ModulType connectedModule = ModulType.UNKNOWN;
	private NetworkCommunication parent;
	private boolean closureAnnounced = false;
	private boolean exit = false;
	
//----------------------- CONSTRUCTORS ----------------------//
	public TCPClientListener(DataStorage dataStorage,
							 HashMap<ModulType, ArrayBlockingQueue<Message> > msgStorage,
							 NetworkCommunication parent,
							 Socket tcpSocket, StreamID streamId) 
	{
		this.dataStorage = dataStorage;
		this.msgStorage = msgStorage;
		this.parent = parent;
		this.tcpSocket = tcpSocket;
		this.streamID = streamId;
	}
	
	public TCPClientListener(DataStorage dataStorage,
			 HashMap<ModulType, ArrayBlockingQueue<Message> > msgStorage,
			 NetworkCommunication parent,
			 Socket tcpSocket, StreamID streamId,
			 ModulType connectedModule) 
	{
		this.dataStorage = dataStorage;
		this.msgStorage = msgStorage;
		this.parent = parent;
		this.tcpSocket = tcpSocket;
		this.streamID = streamId;
		this.connectedModule = connectedModule;
	}
//-------------------- MAINTENANCE METHODS ---------------------//	
	public Socket tcpSocket() {
		return tcpSocket;
	}
	
	@Override
	public void exit() {
		exit = true;
		try {
			tcpSocket.close();
			if (thread != null)
				thread.interrupt();
		} catch (IOException e) {}
	}

//---------------------- RUNABLE METHOD ------------------------//;
	@Override
	public void run() {
		
		thread = Thread.currentThread();
		
		while (!exit) {
			// if in & out is closed, exit
			if (tcpSocket.isInputShutdown() && tcpSocket.isOutputShutdown()) {
				dataStorage.returnUsedStreamID(streamID);
				exit();
			}
			
			Message msg;
			
			if (! tcpSocket.isInputShutdown()) {
				try {
					InputStream in = tcpSocket.getInputStream();
					msg = readMessageFromStream(in);
					if (msg == null) continue;
									
				} catch (IOException e) {
					continue;
				}
			}
			else {
				if (closureAnnounced) {
					try {
						Thread.sleep(TIMEOUT);
					} catch (InterruptedException e) {
						continue;
					}
				}
				
				msg = Message.streamClosedMessage(connectedModule);
				if (msg == null) continue;
				closureAnnounced = true;
			}
		
			msg.streamID = streamID;
			msg.remotePeer = new PeerInfo(msg.sourceID, tcpSocket.getInetAddress(), tcpSocket.getPort());
			
			//dataStorage.debugMessage("TCPListener: processing message: \n" + msg.toString());
			processMessage(msg);
			
			if (connectedModule == ModulType.UNKNOWN)
				connectedModule = msg.targetModule();
		}
		
		try {
			tcpSocket.close();
		} catch (IOException e) {
			dataStorage.raiseError("IO exception while closing TCP socket");
		}
		parent.removeListener(this.streamID);
	}
//--------------------- AUXILIARY -----------------------//
	private Message readMessageFromStream(InputStream in) throws IOException {
		
		//dataStorage.debugMessage("TCPListener: reading from stream (" +in.available()+ " bytes available)");

		Message msg;
		
		byte[] buf = new byte[Message.HEADER_SIZE];
		int len = in.read(buf, 0, buf.length);
		//dataStorage.debugMessage("TCPListener: readed "+len+" bytes from stream (message header)");
		if (len == -1) {
			dataStorage.debugMessage("TCPListener: end of stream reached");
			tcpSocket.shutdownInput();
			return null;
		}
		if (len != buf.length) {
			dataStorage.debugMessage("TCPListener: message length problem");
			return null;
		}
		
		msg = Message.messageHeaderFromBytes(buf);
		if (msg == null) {
			dataStorage.debugMessage("TCPListener: message header from bytes failed");
			return null;
		}
		if (msg.messageType == 0x00 || msg.messageType == 0x01 || 
				msg.messageType == 0x02 || msg.messageType == 0x03){
			dataStorage.debugMessage("TCPListener: wrong message type");
			return null;
		}
			
		msg.data = new byte[msg.dataLength];
		int to_read = msg.dataLength;
		while(to_read > 0){
			len = in.read(msg.data, msg.dataLength - to_read, to_read);
			to_read -= len;
		}
		
		return msg;
	}

}
