package terminalui;

import java.util.ArrayList;

import modules.DataStorage;
import modules.DownloadModul;
import modules.DownloadedFilesDatabase;
import terminalui.TerminalIO.FSColor;
import terminalui.TerminalIO.FSInput;
import terminalui.TerminalIO.FSInputContainer;
import types.DownloadedFileDescriptor;
import types.SearchResult;
import types.SharedFileDescriptor;

public class DownloadCommand extends Command implements FSInputContainer {

	@Override
	public CmdReturn execute(String[] args) {
		String[] supportedFlags = {"status", "help", "quit"};
		if (!parseArgs(args, supportedFlags)) return CmdReturn.INVALID_SYNTAX;
		
		if (flags.contains("help")) return displayHelp();
		if (flags.contains("status")) return displayStatus();
		if (flags.contains("quit")) return quitDownload();
		
		return downloadItems();
	}

	@Override
	public void help() {
		TEXT.out("    Used for downloading files via TazUP.\n");
		TEXT.out("USAGE\n"); 
		TEXT.out("    download [<subcommand>] [<index> [...]]\n");
		TEXT.out("SUBCOMMANDS\n");
		TEXT.out("    help     displays help for this command\n");
		TEXT.out("    status   displays download status\n");
		TEXT.out("    quit     ends downloads specified by indices\n");
		TEXT.out("\n    Without subcommand initiates download of search results specified.\n");
		TEXT.out("INDICES\n");
		TEXT.out("    Indices (No.) in list coresponding to command.");
		TEXT.out("OPTIONS\n");
		TEXT.out("    none\n");
	}
	
	private CmdReturn downloadItems() {
		if (this.args.length == 0) return CmdReturn.MISSING_ARGUMENT;
		
		SearchResult[] results = DataStorage.getInstance().getSearchResultsSortedByTime();
		DownloadModul dm = DataStorage.getInstance().getDownloadInstance();
		
		for (String s: args) {
			int index;
			try { index = Integer.parseInt(s);
			} catch (Exception e) { 
				TEXT.out(s + ": not valid index\n");
				continue; 
			}
	
			if (index < 0 || index >= results.length) {
				TEXT.out(s + ": index out of bounds\n");
				continue;
			}
			
			if (dm.startDownloadFile(results[index]))
				TEXT.out("initiated download of file " + results[index].filename() + "\n");
			else
				TEXT.out("unable to initiate download of file " + results[index].filename() + "\n");
		}
		
		return CmdReturn.SUCCESS;
	}

	private CmdReturn displayStatus() {
		if (args.length != 0) return CmdReturn.INVALID_SYNTAX;
		
		startFullScreen(this);
		
		return CmdReturn.SUCCESS;
	}

	private CmdReturn quitDownload() {
		if (this.args.length == 0) return CmdReturn.MISSING_ARGUMENT;
		
		DownloadModul dm = DataStorage.getInstance().getDownloadInstance();
		DownloadedFilesDatabase db = DataStorage.getInstance().DownloadedFiles();
		
		DownloadedFileDescriptor[] desc = new DownloadedFileDescriptor[db.size()];
		int i=0;
		for (SharedFileDescriptor s: db.getFiles()){
			desc[i] = (DownloadedFileDescriptor) s;
			i++;
		}
		
		for (String s: args) {
			int index;
			try { index = Integer.parseInt(s);
			} catch (Exception e) { 
				TEXT.out(s + ": not valid index\n");
				continue; 
			}
	
			if (index < 0 || index >= desc.length) {
				TEXT.out(s + ": index out of bounds\n");
				continue;
			}
			
			dm.finishDownload(desc[index]);
			TEXT.out("download of " + desc[index].getFilename() + " ended\n");
		}
		
		return CmdReturn.SUCCESS;
	}
	
	private FSInput[] input;
	@Override
	public FSInput[] getData() {
				
		ArrayList<FSInput> in = new ArrayList<FSInput>();
		
		DownloadedFilesDatabase db = DataStorage.getInstance().DownloadedFiles();
		
		//
		in.add(new FSInput("Download status", FSColor.GREEN, 0, 0));
		
		in.add(new FSInput("No.", FSColor.SYSTEM, 0, 2));
		in.add(new FSInput("File name", FSColor.SYSTEM, 10, 2));
		in.add(new FSInput("Status", FSColor.SYSTEM, 60, 2));
		
		int line = 3;
		int i = 0;
		for (SharedFileDescriptor sfd: db.getFiles()) {
			DownloadedFileDescriptor dfd = (DownloadedFileDescriptor)sfd;
			in.add(new FSInput("" + i, FSColor.TEXT, 0, line));
			in.add(new FSInput(dfd.getFilename(), FSColor.TEXT, 10, line));
			in.add(new FSInput(String.format("%.2f", dfd.getProgress()*100) + "%", FSColor.TEXT, 60, line));
			line++;
		}
		
		
		//
		input = new FSInput[in.size()];
		input = in.toArray(input);
		return input;
	}
	
	
}
