package terminalui;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;

import modules.DataStorage;
import terminalui.TerminalIO.FSInputContainer;
import types.PeerInfo;

public class PeerCommand extends Command implements FSInputContainer{

	@Override
	public CmdReturn execute(String[] args) {
		String[] supportedFlags = {"-a", "add", "rm", "help"};
		if (!parseArgs(args, supportedFlags)) return CmdReturn.INVALID_SYNTAX;
		
		if (flags.contains("help")) return displayHelp();
		if (flags.contains("add")) return addPeer();
		if (flags.contains("rm") && flags.contains("-a")) return removeAllPeers();
		if (flags.contains("rm")) return removePeer();
		
		return displayPeers();
	}
	
	@Override
	public void help() {
		TEXT.out("    Used for TazUP peers management.\n");
		TEXT.out("USAGE\n"); 
		TEXT.out("    peer [<subcommand>] [<options>] [...]]\n");
		TEXT.out("SUBCOMMANDS\n");
		TEXT.out("    help     displays help for this command\n");
		TEXT.out("    add      add peer with syntax <ip address> <port>\n");
		TEXT.out("    rm       remove inactive peer with syntax [-a] [<ip address> <port>]\n");
		TEXT.out("\n    Without subcommand displays TazUP peers.\n");
		TEXT.out("OPTIONS\n");
		TEXT.out("    -a       all peers\n");
	}
	
	private CmdReturn displayPeers() {
		if (args.length != 0) return CmdReturn.INVALID_SYNTAX;
		
		startFullScreen(this);
		
		return CmdReturn.SUCCESS;
	}
	
	private CmdReturn addPeer() {
		if (args.length < 2) return CmdReturn.MISSING_ARGUMENT;
		if (args.length != 2) return CmdReturn.INVALID_SYNTAX;
		
		InetAddress ip;
		try {
			ip = InetAddress.getByName(args[0]);
		} catch (UnknownHostException e) {
			TEXT.out(args[0] + " is not valid IP address\n");
			return CmdReturn.SUCCESS;
		}
		
		Integer port;
		try {
			port = Integer.parseInt(args[1]);
			if (port > 65537 || port < 0) throw new NumberFormatException();
		}
		catch (Exception e) {
			TEXT.out(args[1] + " is not valid port\n");
			return CmdReturn.SUCCESS;
		}
		
		DataStorage ds = DataStorage.getInstance();
		ds.addPeer(new PeerInfo(null, ip, port));
		
		return CmdReturn.SUCCESS;
	}
	
	private CmdReturn removePeer() {
		if (args.length < 2) return CmdReturn.MISSING_ARGUMENT;
		if (args.length != 2) return CmdReturn.INVALID_SYNTAX;
		
		InetAddress ip;
		try {
			ip = InetAddress.getByName(args[0]);
		} catch (UnknownHostException e) {
			TEXT.out(args[0] + " is not valid IP address\n");
			return CmdReturn.SUCCESS;
		}
		
		Integer port;
		try {
			port = Integer.parseInt(args[1]);
			if (port > 65537 || port < 0) throw new NumberFormatException();
		}
		catch (Exception e) {
			TEXT.out(args[1] + " is not valid port\n");
			return CmdReturn.SUCCESS;
		}
		
		DataStorage ds = DataStorage.getInstance();
		if (!ds.removePeerByAddress(ip, port))
			TEXT.out(args[0] + " " + args[1] + ": no such peer\n");
		
		
		return CmdReturn.SUCCESS;
	}
	
	private CmdReturn removeAllPeers() {
		DataStorage.getInstance().removeAllPeers();
		return CmdReturn.SUCCESS;
	}

	@Override
	public FSInput[] getData() {
		ArrayList<FSInput> in = new ArrayList<FSInput>();
		DataStorage ds = DataStorage.getInstance();
		
		PeerInfo[] peers;
		String headline;
		
		if (flags.contains("-a")){
			headline = "Peers";
			peers = ds.getPeerList();
		}
		else {
			headline = "Active peers";
			peers = ds.getActivePeers();
		}
		
		//
		in.add(new FSInput(headline, FSColor.GREEN, 0, 0));
	
		in.add(new FSInput("Peer ID", FSColor.SYSTEM, 0, 2));
		in.add(new FSInput("IP Address", FSColor.SYSTEM, 20, 2));
		in.add(new FSInput("Port", FSColor.SYSTEM, 40, 2));
		
		int line = 3;
		for (PeerInfo pi: peers) {
			in.add(new FSInput((pi.id() == null) ? "unknown" : pi.id().toString(), FSColor.TEXT, 0, line));
			in.add(new FSInput(pi.ip().getHostAddress(), FSColor.TEXT, 20, line));
			in.add(new FSInput(""+pi.port(), FSColor.TEXT, 40, line));
			line++;
		}
		//
		FSInput[] input = new FSInput[in.size()];
		input = in.toArray(input);
		return input;
	}
	
}
