package terminalui;

import java.io.File;
import java.util.ArrayList;

import auxiliary.SHA1;

import terminalui.TerminalIO.FSInputContainer;
import types.SharedFileDescriptor;

import modules.DataStorage;
import modules.SharedFilesDatabase;

public class ShareCommand extends Command implements FSInputContainer {

	@Override
	public CmdReturn execute(String[] args) {
		String[] supportedFlags = {"add", "rm", "help"};
		if (!parseArgs(args, supportedFlags)) return CmdReturn.INVALID_SYNTAX;
		
		if (flags.contains("help")) return displayHelp();
		if (flags.contains("add")) return addFiles();
		if (flags.contains("rm")) return deleteFiles();
		if (this.args.length > 0) return addFiles();
		
		return showFiles();
	}
	
	@Override
	public void help() {
		TEXT.out("    Used for shared file list management.\n");
		TEXT.out("USAGE\n"); 
		TEXT.out("    share [<subcommand>] [<options>] [<filename> [...]]\n");
		TEXT.out("SUBCOMMANDS\n");
		TEXT.out("    help     displays help for this command\n");
		TEXT.out("    add      adding files for sharing. Default if filename specified.\n");
		TEXT.out("    rm       removing files for sharing\n");
		TEXT.out("\n    Without subcommand and filename displays shared files list.\n");
		TEXT.out("OPTIONS\n");
		TEXT.out("    none\n");
		TEXT.out("FILENAMES\n");
		TEXT.out("    At least one filename must be specified.\n");
		TEXT.out("    When removing, if only filename (not entire path) is specifed, all matching files are removed.\n");
	}
	
	private CmdReturn addFiles() {
		if (args.length == 0) return CmdReturn.MISSING_ARGUMENT;
		
		for (String s: args) {
			File f = new File(s);
			if (!f.canRead()) TEXT.out(s + ": no such file\n");
			else {
				SharedFilesDatabase db = DataStorage.getInstance().SharedFiles();
				SharedFileDescriptor sfd;
				TEXT.out(s + ": computing hash\n");
				SHA1 hash = SHA1.getFileHash(f.getPath());
				if (hash == null) {
					TEXT.out(s + ": error addinig file\n");
					return CmdReturn.SUCCESS;
				}
				sfd = new SharedFileDescriptor(f.getName(), f.getParent(), hash);
				if(db.add(sfd)){
					TEXT.out(s + ": added\n");
				}
				else{
					TEXT.out(s + ": not added (probably already in the database)\n");
				}
			}
		}
		
		return CmdReturn.SUCCESS;
	}
	
	private CmdReturn deleteFiles() {
		if (args.length == 0) return CmdReturn.MISSING_ARGUMENT;
		
		SharedFilesDatabase db = DataStorage.getInstance().SharedFiles();
		
		ArrayList<SharedFileDescriptor> files = new ArrayList<SharedFileDescriptor>();
		
		for (String s: args) {
			if (!s.contains(File.separator))
				for (SharedFileDescriptor sfd: db.filesByFilename(s))
					files.add(sfd);
			else {
				SharedFileDescriptor sfd = db.fileByFilePath(s);
				if (sfd != null) files.add(sfd);
			}
		}
		
		for (SharedFileDescriptor sfd: files)
			db.remove(sfd);
		
		TEXT.out("Removed " + files.size() + " files\n");
		
		return CmdReturn.SUCCESS;
	}
	
	private CmdReturn showFiles() {
		if (args.length != 0) return CmdReturn.INVALID_SYNTAX;
		
		startFullScreen(this);
		
		return CmdReturn.SUCCESS;
	}

	FSInput[] input = null;
	
	@Override
	public FSInput[] getData() {
		
		if (input != null) return input;
		
		SharedFilesDatabase db = DataStorage.getInstance().SharedFiles();
		
		ArrayList<FSInput> in = new ArrayList<FSInput>();
		//
		in.add(new FSInput("Shared Files", FSColor.GREEN, 0, 0));
		in.add(new FSInput("Filename", FSColor.SYSTEM, 0, 2));
		in.add(new FSInput("File path", FSColor.SYSTEM, 20, 2));
		in.add(new FSInput("Hash", FSColor.SYSTEM, 50, 2));
		
		int line = 3;
		for (SharedFileDescriptor sfd: db.getFiles()) {
			in.add(new FSInput(sfd.getFilename(), FSColor.TEXT, 0, line));
			in.add(new FSInput(sfd.getFilepath(), FSColor.TEXT, 20, line));
			in.add(new FSInput(sfd.getTargetChecksum().toString(), FSColor.TEXT, 50, line));
			line++;
		}
		
		//
		input = new FSInput[in.size()];
		input = in.toArray(input);
		return input;
	}
	
}
