package types;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;

import modules.DataStorage;

import auxiliary.Convert;
import auxiliary.SHA1;

//!TODO synchronized
//!TODO -- ensure that each function which returns boolean == false will be handled (as DownloadClose)
/*
 * contained e.g. unfinished downloaded file..
 * after the application exits (or backups), the list of descriptors is stored. 
 * after the download has been succeeded, we don`t need the descriptor any more. 
 * 
 * Takes care for folder + File.separator + filename + ".tfp" //TFP - Tazup file parts
 */
//Now it is implemented so that I can download from multiple peers at once.
//---- It is just necessary to start a download in DownloadModul with a existing DownloadedFileDescriptor

//Inspired with SearchResult class
public class DownloadedFileDescriptor extends SharedFileDescriptor {
	//Size of file parts in bytes
	//--cannot be 4 (DownloadMessage.PartMessage.parse()
	//UDP max 65536 - header size
	public static final int part_size = 65536;
	
	public class Uploader{
		public Uploader(){
			this.peer_id = null;
			this.sequence_id = null;
			this.stream_id = null;
		}
		public Uploader(PeerID peer_id){
			this();
			this.peer_id = peer_id;
		}
		
		public PeerID peer_id;
		public SequenceID sequence_id;
		public StreamID stream_id;
		
		@Override
		public boolean equals(Object o){
			if(o instanceof DownloadedFileDescriptor.Uploader){
				return ((Uploader)o).peer_id.equals(this.peer_id);
			}
			return false;
		}
	}
	
	private long file_size = 0;
	//Signalize between DownloadThreads, that this file is successfully downloaded. 
	private boolean is_completized = false;
	
	//On the beginning it allocates enough size for the downloaded file. 
    private RandomAccessFile out = null;
    
    //Which parts has been already downloaded
    private ArrayList<Boolean> parts = null; 
    private int parts_saved = -1;
    
    //!TODO uploaders manage
    public ArrayList<Uploader> uploaders = null;
    
    public DownloadedFileDescriptor(String filename, String folder, SHA1 target_checksum){
    	super(filename, folder, target_checksum);
    	this.out = null;
    	this.parts = null;
    	this.uploaders = new ArrayList<DownloadedFileDescriptor.Uploader>();
    }
    
    public int getPartsCount(){
    	//Horna cela cast
    	return (int)(
    			(this.file_size + (long)DownloadedFileDescriptor.part_size - (long)1) 
    			/ ((long)DownloadedFileDescriptor.part_size)
    	);
    }
    
    public SHA1 getChecksum(){
    	return this.target_checksum;
    }
    
    //!TODO - Event (for JProgressBar)
    public double getProgress(){
    	if(this.parts == null || this.parts.isEmpty() 
    			|| this.getPartsCount() == 0 || this.parts_saved < 0)
    		return 0.0;
    	return ((double) this.parts_saved) / ((double) this.getPartsCount());
    }
    public int getDownloadedPartsCount(){
    	return this.parts_saved;
    }
    //!TODO -- test new implementation
    public boolean finished(){
    	return (this.parts_saved == this.getPartsCount());
    }
    
    //The parts file filename
    public String getFilePartsFilepath(){
    	return this.getFilepath() + ".tfp";
    }
    //Tazup downloaded parts
    public String getBufferFilepath(){
    	return this.getFilepath() + ".tdp";
    }

	//Open for reading and writing, as with "rw", 
    //--also computes the length
	//--and also require that every update to the file's content or 
	//--metadata be written synchronously to the underlying storage device.
    protected boolean openBufferFile(){
    	//!TODO -- I hope that rws will not write over that file if exists.
    	try {
			this.out = new RandomAccessFile(this.getBufferFilepath(), "rws");
		} catch (FileNotFoundException e) {
			DataStorage.getInstance().debugMessage(this.getBufferFilepath()+ " - cannot open buffer file.");
			return false;
		}
		try {
			this.file_size = this.out.length();
		} catch (IOException e) {
			return false;
		}
		return true;
    }

    //!TODO -- developer only
    public boolean saveBufferFile(){
    	try {
			this.out.close();
		} catch (IOException e) {
			return false;
		}
		return true;
    }
    //!TODO -- developer only
    public ArrayList<Boolean> getFileParts(){
    	return this.parts;
    }
    
    //filename.txt -> filename(2).txt -> filename(3).txt
    //--for example Mozilla uses this convention
    public static String generateNotExistingFilepath(String filepath){
		if(!(new File(filepath)).exists()) 
			return filepath;
		
    	Integer i=2;
    	//There cannot be infinite amount of files allocated
    	while(true){
    		int pos = filepath.lastIndexOf('.');
    		if(pos < 0) pos = filepath.length();
    		String result = filepath.substring(0, pos) + "("+i.toString()+")" + ((pos>=filepath.length())?"":filepath.substring(pos));
    		if(!(new File(result)).exists()) {
    			return result;
    		}
    		
    		i++;
    	}
    }
    
	//O(n)
	public DownloadedFileDescriptor.Uploader getUploaderByPeerIdValue(int value){
		synchronized(this) {
			int index = this.uploaders.indexOf(this.new Uploader(new PeerID(value)));
			if(index < 0)
				return null;
			
			return this.uploaders.get(index);
		}
	}
	
    private void allocateParts(int parts_count){
    	synchronized(this){
	    	this.parts = new ArrayList<Boolean>(parts_count);
	    	for(int i=0; i<parts_count; i++){
	    		//Why the initializer not works I don`t know. 
	    		this.parts.add(false);
	    	}
    	}
    }

    private int last_missing_part = 0;
    //Call just from DownloadThread.run(), 
    //->because it can block the thread for some time when going thru the parts list once more. 
    
    public int getMissingPart(){
    	synchronized(this){
	    	if(this.finished())
	    		this.last_missing_part = -1;
	    	if(this.last_missing_part < 0)
	    		return -1;
	    	
	    	//We don't mind concurrency issues, because the bits are set only from false to true. 
	    	for(int i=this.last_missing_part; i<this.parts.size() ; i++){
	    		if(!this.parts.get(i)){
	    			//Heuristics. 
	    			this.last_missing_part=i+1;
	    			return i;
	    		}
	    	}
	    	//If some part is still false, set last_missing_part to 0
	    	//--remark that this can make to several requests for the same part (so Moduls should ignore old messages)
	    	//--O(n)
	    	if(!this.isSomeMissingPart()){
	    		this.last_missing_part = -1;
	    		return -1;
	    	}
	    	
	    	try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {}
	    	
	    	//!Recursion
	    	this.last_missing_part = 0;
	    	return this.getMissingPart();
    	}
    }
    
    //Goes thru the whole part array and returns if some part is false. 
    //!O(n)
    public boolean isSomeMissingPart(){
    	for(int i=0; i<this.parts.size() ; i++)
    		if(!this.parts.get(i))
    			return true;
    	return false;
    }
    
    //!TODO
    //--Ensure that false response is handled in al function calls.
    public boolean addPart(int part_id, byte[] data){
    	if(data == null){
    		DataStorage.getInstance().debugMessage("Tried to add null part_id="+ part_id + " for file " + this.getFilepath());
    		return false;
    	}
    	if(this.parts.get(part_id)){
    		DataStorage.getInstance().debugMessage("Part " + part_id + " already downloaded.");
    		return true;
    	}
    	
    	//This function just needs to be atomic.
    	//-- for example two addPart function could be initiated. 
    	//-- or part_id is set true, but the count is not (so part=-1 can be requested).
    	synchronized(this){
	    	byte[] d = data;
	    	long offset = ((long)part_id) * ((long)DownloadedFileDescriptor.part_size);
	    	
	    	//Based on test, RandomAccessFile.write(byte[]) 
	    	//--will try to write as much bytes as needed by byte[]

	    	try {
		    	int length = (int)((part_id < this.getPartsCount() - 1) ? part_size : (this.out.length() % (long)part_size));
		    	if(length != part_size){
		    		d = new byte[length];
		    		System.arraycopy(data, 0, d, 0, length);
		    	}
	    	//Import, because some files could be longer then 2MB
				this.out.seek(offset);
				this.out.write(d);
	    	} catch (IOException e) {
	    		DataStorage.getInstance().debugMessage("Unable to write part " + part_id + " of file " + this.getFilepath());
	    		return false;
			}
	    	this.parts.set(part_id, true);
	    	this.parts_saved++;
	    	DataStorage.getInstance().debugMessage("File part " + part_id + " of '"+this.getFilename()+"' is home ("+this.parts_saved + " parts from " + this.getPartsCount() + ")");
    	}
    	return true;
    }
    
    //Allocates file on users file system (file_size in bytes)
    //called when download starts.
    //--also creates a descriptor file (which parts are missing)
	//--if something was downloaded earlier, 
	//----then it tries to load the parts and continue from where it ended. 
    public boolean allocate(long file_size){
    	synchronized(this) {
    		if(this.file_size > 0 && this.out != null && this.parts != null){
    			DataStorage.getInstance().debugMessage("File " + this.getFilepath() + " already allocated.");
    			return true;
    		}
    		
    		this.file_size = file_size;
	    	//Tries to get the saved fileparts. 
	    	//--if not successful, then deleted them and calls allocate once more
	    	if(
	    			(new File(this.getBufferFilepath())).exists()
	    		&&	(new File(this.getFilePartsFilepath())).exists()
	    	){
	    		if(this.load()){
	    			DataStorage.getInstance().debugMessage("Parts successfull loaded: " + this.getBufferFilepath());
	    			return true;
	    		}
	    	}

	    	//Parts or Buffer file is not accessible. 
	    	//Create a file big enough for the file, store it in out. 
	    	DataStorage.getInstance().debugMessage("New buffer file " + this.getBufferFilepath() + " allocated");
	    	try{
	    		RandomAccessFile raf = new RandomAccessFile(this.getBufferFilepath(), "rw");
	    		raf.setLength(file_size);
	    		raf.close();
	    	}catch (FileNotFoundException e){ //I cannot even imagine, that this error occurs (rw). 
	    		System.out.println("File: " + this.getBufferFilepath() + " not exists at DFD.alocate()");
	    		return false;
	    	}catch (IOException e){
	    		System.out.println("File: " + this.getBufferFilepath() + " cannote be prepared at DFD.alocate()");
	    		return false;
	    	}

	    	this.file_size = file_size;
	    	this.parts_saved = 0; 

	    	this.last_missing_part = 0;
	    	//Allocate and Backup parts
	    	int parts_count = this.getPartsCount();
	    	this.allocateParts(parts_count);

	    	this.openBufferFile();

    	}
    	return true;
    }
    
    //Stores the target file to target position
    //--deletes all not necessary files (parts file and buffer file)
    public boolean completize(){
    	synchronized(this) {
    		if(this.is_completized){
    			DataStorage.getInstance().debugMessage("File " + this.getFilepath() + " already completized.");
    			return true;
    		}
	    	if(!this.finished()){
	    		DataStorage.getInstance().debugMessage("Cannot completize download of file " + this.getFilepath() + " because the download is not yet completed.");
	    		return false;
	    	}
	    	
	    	try {
				this.out.close();
			} catch (IOException e) {
				DataStorage.getInstance().debugMessage("Buffer file cannot be saved/closed: " + this.getBufferFilepath());
				return false;
			}
			
	    	//Rename the target file
	    	File f1 = new File(this.getBufferFilepath());
	    	
	    	String toSave = DownloadedFileDescriptor.generateNotExistingFilepath(this.getFilepath());
	    	File f2 = new File(toSave);
	    	if(!f1.renameTo(f2)){
	    		DataStorage.getInstance().debugMessage("DownloadedFileDescriptor of " + this.getFilepath() + " was unable to rename '"+this.getBufferFilepath()+"' to '"+toSave+"'");
	    		return false;
	    	}
	
			SHA1 sha = null; 
			try {
				sha = SHA1.getFileHash(toSave);
			} catch (Exception e) {}
			
			boolean error = false; 
			if(!this.getTargetChecksum().equals(sha)){
				error = true;
				DataStorage.getInstance().debugMessage("Checksums not correspond for file=" + this.getFilepath() + "(\ntarget="+this.getTargetChecksum()+"\nfile  ="+sha+"\n)");
			}
	    	
	    	//Remove the parts file
			//--even if the checksum not correspond (it is impossible to find the part in which is the error)
	    	File f = new File(this.getFilePartsFilepath());
	    	if(!f.delete()){
	    		DataStorage.getInstance().debugMessage("Parts file cannot be deleted: " + this.getFilePartsFilepath());
	    		return false;
	    	}
	    	
	    	if(!error){
	    		this.is_completized = true;
	    	}
	    		
	    	return !error;
		}
    }
    
    //Stores the file parts file, saves the output file
    // -- the try / catch blocks made this functions code quite unreadable...
    //!TODO save also peer ids
    public boolean backup(){
    	//No other thread can set part / write part why I store them.  
    	synchronized (this) {
    		if(this.parts == null){
    			DataStorage.getInstance().debugMessage("Parts are null, cannot backup " + this.getFilepath());
    			return false;
    		}
    		
        	//Buffer file is backup-ed synchronously with writing into it. 
        	FileOutputStream fos;
        	
			try {
				fos = new FileOutputStream(this.getFilePartsFilepath());
			} catch (FileNotFoundException e) {
				DataStorage.getInstance().debugMessage(this.getFilePartsFilepath() + " - cannot open parts for write.");
				return false;
			}
			//Save downloaded parts to file handler. 
	    	for(int i=0; i<this.parts.size();i++){
	    		if(this.parts.get(i)){
	    			try {
						fos.write(Convert.intToBytes(i));
					} catch (IOException e) {
						DataStorage.getInstance().debugMessage(this.getFilePartsFilepath() + " - cannot write part file.");
						return false;
					}
	    		}
	    	}
	    	//Save downloadedBufferFile
	    	try {
				this.out.close();
			} catch (IOException e) {
				DataStorage.getInstance().debugMessage(this.getBufferFilepath() + " - cannot close(save) buffered file.");
				return false;
			}
	    	if(!this.openBufferFile()){
				DataStorage.getInstance().debugMessage(this.getBufferFilepath() + " - cannot reopen buffered file.");
				return false;
	    	}
	    	//Save partsFile
	    	try {
				fos.close();
			} catch (IOException e) {
				DataStorage.getInstance().debugMessage(this.getFilePartsFilepath() + " - cannot close(save) parts file.");
				return false;
			}
    	}
		return true;
    }
    

	//Open FilePartsFile and set those parts as already downloaded. 
    //!also needs bufferedFile to be opened
    //should be called only from load() 
    //!TODO -- test implementation
    protected boolean loadParts(){
    	synchronized(this){
	    	//Test if all is well on the FileSystem side.
	    	if(this.out == null)
	    		if(!this.openBufferFile()){
	    			DataStorage.getInstance().debugMessage("DFD.allocate(): cannot open buffer file " + this.getBufferFilepath());
	    			return false;
	    		}
	    	if(this.file_size <= 0){
	    		DataStorage.getInstance().debugMessage("DFD.allocate(): file_size is small: " + this.file_size);
	    		return false;
	    	}
	    	//openBufferFile() also computes the file size
	    	
	    	RandomAccessFile in = null;
			try {
				 in = new RandomAccessFile(this.getFilePartsFilepath(), "r");
			} catch (FileNotFoundException e) {
				//No parts file exists, so nothing to do. 
				//--it will be created in allocate method hopefully
				DataStorage.getInstance().debugMessage("Cannot open fileParts file: " + this.getBufferFilepath());
				return false;
			}
			
			byte[] data = null;
			try {
				data = new byte[(int)in.length()];
				in.read(data);
			} catch (IOException e) {
				DataStorage.getInstance().debugMessage(this.getFilepath() + " - cannot read data.");
				return false;
			}
			
			if(data.length % 4 != 0){
				DataStorage.getInstance().debugMessage(this.getFilepath() + " - corrupted parts_count.");
				return false;
			}
	
			this.allocateParts(this.getPartsCount());
			
			boolean error = false;
			for(int i=0; i<data.length ; i+=4){
				int index = Convert.bytesToInt(data, i);
				if(index < 0 || index >= this.parts.size()){
					DataStorage.getInstance().debugMessage(this.getFilepath() + " - part index out of range (at i="+i+" index="+index+" parts_count="+this.parts.size()+").");
					error=true;
					continue;
				}
				this.parts.set(index, true);
			}
			
			//Just to retain "some consistency"
			if(!error){
				this.parts_saved = data.length / 4;
			}
			return !error;
    	}
    }
    
    //Can be called directly after constructor
    //--PeerIds will be saved in the DownloadedFilesDatabase
    public boolean load(){
    	if(!this.openBufferFile()){
    		DataStorage.getInstance().debugMessage("Cannot openBufferFile in DFD.load() for file: " + this.getFilepath());
    		return false;
    	}
        if(!this.loadParts()){
        	DataStorage.getInstance().debugMessage("Cannot loadParts in DFD.load() for file: " + this.getFilepath());
        	return false;
        }
        	
    	return true;
    }
               
}
