package types;

import modules.DataStorage;
import auxiliary.Convert;

/*
 * used for communicaton between modules
 * data - content of packet (withouth lower protocols headers)
 * targetPeer - reciever of message
 * targetModule - should be consistent with data, ignored while sending message
 * streamID - each TCP transfer should have unique streamID. Thus when answering on message, module should copy/paste, 
 *            when iniciating TCP transfer, module should generate new unique ID, through dataStorage.generateNewStreamID()
 */

/*
 * ============= PROTOCOL DESCRIPTION =============
 * BITS		MEANING
 * 0-23		protocol identifier
 * 24-31	message type
 * 32-63	source ID
 * 64-95	target ID
 * 96-127	identifier
 * 128-159	data length
 * 160-..	data	
 * 
 * protocol identifier
 * ===================
 * should be "tzp" 
 * 
 * message type
 * ============
 * describes purpose of message:
 * 
 * UDP transport:
 * 0x00 - tazup ping
 * 0x01 - tazup pong
 * 0x02 - search query
 * 0x03 - search result
 * 
 * TCP transport
 * 0x04 - download init (download / upload is resolved by target module)
 * 0x05 - download request part
 * 0x06 - download cancel
 * 
 * total length
 * ============
 * length of header + data
 * 
 * source ID
 * =========
 * TazUp ID of sender node
 * 
 * target ID
 * =========
 * Tazup ID of reciever node
 * 
 * identifier
 * ==========
 * has different meanings depending on message type, usualy identifies query/answer pairs
 * 
 * ping/pong - valid replay must have identical identifier
 *             identifier = 0     => an ad hoc ping to a new seen user
 *             identifier = n > 0 => keep-alive ping sent to the n-th peer in the current actualPeers[] array
 * search query/answer - answer/forwarded query has the same identifier as initial query
 * upload/download - all messages in single file transfer has the same identifier
 * 
 * data length
 * ===========
 * length of data in bytes
 * 
 * data
 * ====
 * content depends on message type:
 * 
 * PING / PONG: random data, should be the same in answer
 * 
 * SEARCH QUERY: 
 * 		//TODO port size should be only 2 bytes...
 * 		initiator info:
 * 		bits
 * 		0-31	ttl
 * 		32-63	initiator ID
 * 		64-95	initiator IP
 * 		96-127	initiator port
 * 		128-..	UTF-8 search string: keywords separated by space
 * 
 * SEARCH RESULT:
 * 		bits
 * 		0-159 	SHA1(file)
 * 		160-167	filesize
 * 		168-..	filename (UTF-8) 
 * 
 * DOWNLOAD REQUEST:
 * 		(More in DownloadMessage)
 * 		ASCII string in format:
 * 		filename;filesize;hash;starting_byte\n
 * 
 * FILE UPLOAD:
 * 		bits
 * 		0-31		data offset in bytes
 * 		32-(of-1)	info string (starting byte, file info)
 * 		off-..		binary data
 * 
 * UPLOAD CANCELED: single byte, containing error description:
 * 		0x00: unspecified error
 * 		0x01: no such file
 */

public class Message {
	static public int HEADER_SIZE = 20;
	
	// non protocol, messageFromBytes leaves them on null
	public StreamID streamID; 		
	public boolean streamClosed; 	// when sending -> i wanna close connection, recieving -> he closed connection
	public PeerInfo remotePeer;
	// protocol 
	public byte messageType;
	public PeerID sourceID;		
	public PeerID targetID;
	public SequenceID identifier;
	//!TODO -- redundancy
	public int dataLength;
	public byte[] data;
	
	//With deriving this class you can make more human-friendly byte[] data representation. 
	//--some examples can be found in Download/Upload Module
	public interface MessageDataRepresentation{
		boolean parse(byte[] data);
		byte[] encode();
		byte getMessageType();
	}
	
	public MessageDataRepresentation representation = null; 
	
	public void encode(){
		if(this.representation != null){
			this.data = this.representation.encode();
			this.dataLength = this.data.length;
			this.messageType = this.representation.getMessageType();
		}
	}
	
	public ModulType targetModule() {
		switch (messageType) {
			case 0x00: 
			case 0x01: return ModulType.PEER_MANAGEMENT;
			case 0x02: return ModulType.SEARCH_ANSWER;
			case 0x03: return ModulType.SEARCH;
			case 0x04: 
			case 0x05:
			case 0x06: return ModulType.DOWNLOAD;
			case 0x14: 
			case 0x15:
			case 0x16: return ModulType.UPLOAD;
			
			default: return ModulType.UNKNOWN;
		}
	}
	
	public byte[] getBytes() {
		
		if (sourceID == null) return null;
		if (targetID == null) return null;
		if (identifier == null) return null;
		if (data == null) return null;
		
		byte[] srcID = sourceID.getBytes();
		byte[] trgID = targetID.getBytes();
		byte[] seqID = identifier.getBytes();
		byte[] dataLen = Convert.intToBytes(dataLength);
		byte[] bytes = new byte[data.length + HEADER_SIZE];
		
		bytes[0] = 't';
		bytes[1] = 'z';
		bytes[2] = 'p';
		bytes[3] = messageType;
		for (int i=0; i<4; i++) bytes[4+i] = srcID[i];
		for (int i=0; i<4; i++) bytes[8+i] = trgID[i];
		for (int i=0; i<4; i++) bytes[12+i] = seqID[i];
		for (int i=0; i<4; i++) bytes[16+i] = dataLen[i];
		for (int i=0; i<data.length; i++) bytes[HEADER_SIZE+i] = data[i];
		
		return bytes;
	}
	
	public static Message messageHeaderFromBytes(byte[] bytes) {
		if (bytes == null) {
			DataStorage.getInstance().debugMessage("null bytes");
			return null;
		}
		if (bytes[0] != 't' || bytes[1] != 'z' || bytes[2] !='p') {
			DataStorage.getInstance().debugMessage("bad header");
			return null;
		}
		if (bytes.length < HEADER_SIZE) {
			DataStorage.getInstance().debugMessage("small bytes length");
			return null;
		}
		
		Message msg = new Message();
		msg.messageType = bytes[3];
		msg.sourceID = PeerID.peerIDfromBytes(bytes, 4);
		msg.targetID = PeerID.peerIDfromBytes(bytes, 8);
		msg.identifier = SequenceID.sequenceIDfromBytes(bytes, 12);
		msg.dataLength = Convert.bytesToInt(bytes, 16);
		
		return msg;
	}
	
	public static Message messageHeaderFromMessage(Message msg) {
		
		if (msg == null) return null;
		
		Message msgh = new Message();
		msgh.remotePeer = msg.remotePeer;
		msgh.messageType = msg.messageType;
		msgh.sourceID = msg.sourceID;
		msgh.targetID = msg.targetID;
		msgh.identifier = msg.identifier;
		msgh.dataLength = msg.dataLength;
		
		return msgh;
	}
	
	public static Message messageFromBytes(byte[] bytes, int len) {
		Message msg = messageHeaderFromBytes(bytes);
		if (msg == null) return null;
		if (len > bytes.length) {
			DataStorage.getInstance().debugMessage("expected length not matched");
			return null;
		}
		if (msg.dataLength < 0) {
			DataStorage.getInstance().debugMessage("impossible error");
			return null;
		}
		if (msg.dataLength != len - HEADER_SIZE) {
			DataStorage.getInstance().debugMessage("bad length case last");
			return null;
		}
		
		msg.data = new byte[msg.dataLength];
		for (int i=0; i<msg.dataLength; i++)
			msg.data[i] = bytes[HEADER_SIZE + i];
		
		return msg;
	}
	
	public static Message streamClosedMessage(ModulType module) {
		Message msg = new Message();
		msg.streamClosed = true;
		switch (module) {
			case PEER_MANAGEMENT: msg.messageType = 0x01;
				break;
			case SEARCH: msg.messageType = 0x03;
				break;
			case SEARCH_ANSWER: msg.messageType = 0x02;
				break;
			case DOWNLOAD: msg.messageType = 0x06;
				break;
			case UPLOAD: msg.messageType = 0x04;
				break;
			case UNKNOWN: msg.messageType = 0x07;
				break;
		}
		return msg;
	}
        
	//Note: Representation will be copied just as reference.
        public Message deepCopy() {
                Message copy = new Message();
                
                if (streamID == null) copy.streamID = null;
                else copy.streamID =  new StreamID(this.streamID.value());
                
                copy.streamClosed = this.streamClosed;
                
                if (remotePeer == null) copy.remotePeer = null;
                else copy.remotePeer = new PeerInfo(this.remotePeer.id(), this.remotePeer.ip(), this.remotePeer.port());
                
                copy.messageType = this.messageType;
                
                if (sourceID == null) copy.sourceID = null;
                else copy.sourceID = new PeerID(this.sourceID.value());
                
                if (targetID == null) copy.targetID = null;
                else copy.targetID = new PeerID(this.targetID.value());
                
                if (identifier == null) copy.identifier = null;
                else copy.identifier = new SequenceID(this.identifier.value());
                
                copy.dataLength = this.dataLength;
                
                //See function Note. 
                copy.representation = this.representation;
                
                if (data == null) copy.data = null;
                else copy.data = this.data.clone();
                
                return copy;
        }
        
        public Message getDefaultReply(){
        	DataStorage data_storage = DataStorage.getInstance();
        	PeerID peer_id;
        	Message result = this.deepCopy();
        	peer_id = result.sourceID;
        	result.sourceID = result.targetID;
        	result.targetID = peer_id;
        	result.remotePeer = data_storage.findPeerByID(targetID);
        	
        	return result;
        }
        
        public Message getReply(byte[] data) {
        	Message result = this.getDefaultReply();
        	result.data = data;
        	result.dataLength = result.data.length;
        	return result;
        }
        
        @Override
        public String toString(){
        	//public StreamID streamID; 		
        	//public boolean streamClosed; 	// when sending -> i wanna close connection, recieving -> he closed connection
        	//public PeerInfo remotePeer;
        	// protocol 
        	//public byte messageType;
        	//public PeerID sourceID;		
        	//public PeerID targetID;
        	//public SequenceID identifier;
        	//public int dataLength;
        	//public byte[] data;
        	
        	return 
        	"Modul       : " + this.targetModule().toString() + "\n" + 
        	"StreamID    : " + ((this.streamID!=null)?this.streamID.value():"null") + " \n" +
        	"StreamClosed: " + this.streamClosed + "\n" +
        	"RemotePeer  : " + ((this.remotePeer!=null)?this.remotePeer.toString():"null") + "\n" + 
        	"SequenceID  : " + ((this.identifier!=null)?this.identifier.toString():"null") + "\n" +	
        	"MessageType : 0x" + Convert.byteToAlphaHex(this.messageType) + "\n" +
        	"SourceId    : " + ((this.sourceID!=null)?this.sourceID.value():"null") + "\n" + 
        	"TargetId    : " +((this.targetID!=null)?this.targetID.value():"null") + "\n" +
        	"DataLength  : " + this.dataLength + "\n" + 
        	"Data(Hex)   : " + ((this.data!=null)?Convert.bytesToAlphaHexString(this.data):"null") + "\n"
        	;
        }
}
