package types;

import java.net.InetAddress;
import java.lang.Comparable;
import java.lang.ClassCastException;

/** === PeerInfo class ===                    **/
/** contains a detailed description of a peer **/

/** IMPORTANT NOTE!!! **/
/** In case a peer disconnects, reconnects and generates a new random ID **/
/** the "equals" function (and some other classes such as PeerList) will **/
/** consider him a DIFFERENT peer. Therefore, If a peer receives         **/
/** communication intended to his address + port but different peerID    **/
/** he should NOT answer it (or he will confuse the sender). Instead, he **/
/** should react with a message that a new peer (himself) has appeared,  **/
/** thus informing the sender about his new peerID.                      **/

/** IMPORTANT NOTE!!!                                                    **/
/** If handling PeerInfo class, don't attempt to access the "id" member  **/
/** without checking if it isn't "null" first. It could be and WILL BE   **/
/** null on many occasions - because ids are not saved to the file.      **/


public class PeerInfo implements Comparable<Object> {
        private PeerID id;
        private InetAddress ip;
        private int port;
        
        /* constructor */
        public PeerInfo(PeerID id, InetAddress ip, int port) {
                this.id = id;
                this.ip = ip;
                this.port = port;
        }
        
        /* read methods */
        public PeerID id() { return id; }
        public InetAddress ip() { return ip; }
        public int port() { return port; }
        
        /* general comparator */
        @Override
        public boolean equals(Object obj) {
                if (! (obj instanceof PeerInfo)) return false;
                PeerInfo pinfo = (PeerInfo) obj;
                
                return (((this.id == null && pinfo.id() == null) || this.id.equals(pinfo.id())) &&
                        this.ip.equals(pinfo.ip()) &&
                        this.port == pinfo.port());
        }
        
        /* comparator for sets */
        @Override
        public int compareTo(Object obj) {
                if (! (obj instanceof PeerInfo) )
                        throw new ClassCastException();
                
                // ...if both are not null
                if ((this.id != null) && (((PeerInfo) obj).id() != null)) {
                        if (this.id.value() < (((PeerInfo) obj).id().value()))
                                return -1;
                        if (this.id.value() > (((PeerInfo) obj).id().value()))
                                return 1;
                }
                
                // ...if exactly one is not null
                if ((this.id != null) != (((PeerInfo) obj).id() != null)) {
                        if (this.id == null)
                                return -1;
                        else
                                return 1;
                }
                
                if (! this.ip.equals(((PeerInfo) obj).ip())) {
                        byte[] myBytes = this.ip.getAddress();
                        byte[] theirBytes = ((PeerInfo) obj).ip().getAddress();

                        for (int i = 0; i < 4; i++)
                                if (myBytes[i] != theirBytes[i])
                                        return (int)myBytes[i] - (int)theirBytes[i];
                }
                
                if (this.port < ((PeerInfo) obj).port()) return -1;
                if (this.port > ((PeerInfo) obj).port()) return 1;
                
                return 0;
        }
        @Override
        public String toString(){
        	return "PeerInfo id: " + ((this.id == null) ? "null" : this.id.value ) + 
        	" ip: " + ((ip  == null) ? "null" : ip.getHostAddress()) 
        	+ " port: " + port;
        }
        
}
