package types;

import java.io.File;
import java.lang.Comparable;
import java.lang.ClassCastException;
import java.nio.charset.Charset;
import java.util.ArrayList;

import auxiliary.Convert;
import auxiliary.SHA1;

/*
 * single search result (e.g. filename-peer)
 * more of them are usually associated with one search initiated by user
 * 
 * used also for download initiation
 */

public class SearchResult implements Comparable<Object> {
        /* private data */
        private PeerID peer;
        private String filename;
        //!TODO not used in DownloadModul 
        //--(the filesize is an important info, but it is agreed in the init phase)
        private long filesize;
        private SHA1 checksum;
        
        /* constructor */
        public SearchResult(PeerID peer, String filename, long filesize, SHA1 checksum) {
                this.peer = peer;
                this.filename = filename;
                this.filesize = filesize;
                this.checksum = checksum;
        }
        
        /* copy constructor */
        public SearchResult(SearchResult srt) {
                synchronized (srt) {
                        this.peer = srt.peer();
                        this.filename = srt.filename();
                        this.filesize = srt.filesize();
                        this.checksum = srt.checksum();
                }
        }
        
        /* get methods */
        public PeerID peer() { return peer; }
        public String filename() { return filename; }
        public long filesize() { return filesize; }
        public SHA1 checksum() { return checksum; }
            
        
        /* comparator of identity */
        @Override
        public boolean equals(Object obj) {
                if (! (obj instanceof SearchResult)) return false;
                SearchResult srt = (SearchResult) obj;
            
                return (this.peer.equals(srt.peer()) &&
                        this.filename.equals(srt.filename()) &&
                        this.filesize == srt.filesize() &&
                        this.checksum.equals(srt.checksum()));
        }
        
        /* comparation of source (i.e. same PeerID) */
        public boolean equalSource(SearchResult srt) {
                return (this.peer.equals(srt.peer()));
        }
        
        /* comparation of name (same filename, not necessarily checksum!) */
        public boolean equalName(SearchResult srt) {
            return (this.filename.equals(srt.filename()));
        }
        
        /* comparation of file (same filename & filesize & checksum) */
        public boolean equalFile(SearchResult srt) {
                return (this.filename.equals(srt.filename()) &&
                        this.filesize == srt.filesize() &&
                        this.checksum.equals(srt.checksum()));
        }
        
        /* comparator - used to sort by filename & checksum */
        public int compareTo(Object obj) {
                if (! (obj instanceof SearchResult) )
                        throw new ClassCastException();
            
                if (this.filename.compareTo(((SearchResult) obj).filename()) != 0)
                        return this.filename.compareTo(((SearchResult) obj).filename());
                
                if (this.filesize < ((SearchResult) obj).filesize()) return -1;
                if (this.filesize > ((SearchResult) obj).filesize()) return 1;
                
                if (this.checksum.compareTo(((SearchResult) obj).checksum()) != 0)
                        return (this.checksum.compareTo(((SearchResult) obj).checksum()));
                
                if (this.peer.value() < ((SearchResult) obj).peer().value())
                        return -1;
                if (this.peer.value() > ((SearchResult) obj).peer().value())
                        return 1;
                
                return 0;
        }
        
        /* used for creating answer Message */
    	public byte[] getBytes() {
    		if (filename == null) return null;
    		if (checksum == null) return null;
    		
    		ArrayList<Byte> bytes = new ArrayList<Byte>();
    		
    		for (byte b: checksum.value()) bytes.add(b);
    		
    		byte[] sizeb = Convert.longToBytes(filesize);
    		for (byte b: sizeb) bytes.add(b);
    		
    		for (byte b: filename.getBytes(Charset.forName("UTF-8"))) bytes.add(b);
    		
    		byte[] result = new byte[bytes.size()];
    		for (int i=0; i<bytes.size(); i++)
    			result[i] = bytes.get(i);
    		
    		return result;
    	}
        
        /* used for retrieving it from message */
        public static SearchResult fromMessage(Message msg) {
        	if (msg == null) return null;
        	if (msg.messageType != 0x03) return null;
        	if (msg.sourceID == null) return null;
        	if (msg.data.length < SHA1.SHA1_LENGTH + 8 + 1) return null; // checksum + filesize + nonempty filename
        	
        	byte[] checksum = new byte[SHA1.SHA1_LENGTH];
        	for (int i=0; i<SHA1.SHA1_LENGTH; i++) checksum[i] = msg.data[i];
        	SHA1 sha1 = new SHA1(checksum);
        	
        	long filesize = Convert.bytesToLong(msg.data, SHA1.SHA1_LENGTH);
        	
        	byte[] fnm = new byte[msg.data.length - SHA1.SHA1_LENGTH - 8];
        	for (int i=SHA1.SHA1_LENGTH + 8; i<msg.data.length; i++)
        		fnm[i - SHA1.SHA1_LENGTH - 8] = msg.data[i];
        	String filename = new String(fnm, Charset.forName("UTF-8"));
        	
        	return new SearchResult(msg.sourceID, filename, filesize, sha1);	
        }
        
        /* used for sending answer */
        public static SearchResult fromSharedFileDescriptor(SharedFileDescriptor sfd) {
        	
        	File f = new File(sfd.getFilepath());
    		if (!f.canRead()) return null;
    		long filesize = f.length();
    		
    		return new SearchResult(null, sfd.getFilename(), filesize, sfd.getTargetChecksum());
        }
        
      
}
