package types;

import java.io.File;

import modules.DataStorage;
import auxiliary.SHA1;

/* Peto, 24.04.2011
 * Stores data which is needed to share a file from file system. 
 * Provides some basic functionality.
 * Used in 
 * --UploadThread (opens the file and uploads it by parts)
 * --parent class for DownloadedFileDescriptor (also manages the storage of the parts)
 * */
//Backup is executed in SharedFilesDatabase

public class SharedFileDescriptor implements Comparable<Object> {
	//Filename can be searched, if some peer has been lost. 
    protected String filename;
    
    //Where the user wants to store this file 
    //-- /root/subfolder/folder
    //---- !Operating system dependent
    // --- folder=null equiv to folder = ""
    protected String folder; 
    
    //Checksum which will be checked after the download has been finished.
    protected SHA1 target_checksum;
    
    public static SharedFileDescriptor createDescriptorFromFilepath(String filepath){
    	if(!(new File(filepath)).exists()){
    		DataStorage.getInstance().debugMessage("File " + filepath + " is no longer shared.");
    		return null;
    	}
    	
        int pos = filepath.lastIndexOf(File.separator);
        String filename = filepath.substring(pos + 1);
        //without separator
        String folder = filepath.substring(0, pos);
        SHA1 target_checksum = SHA1.getFileHash(filepath);
        	
        if(folder == null) folder = "";
        if(filename == null) filename = "";
        
        return new SharedFileDescriptor(filename, folder, target_checksum);
    }
    
    public SharedFileDescriptor(String filename, String folder, SHA1 target_checksum){
    	this.filename = filename;
    	this.folder = folder;
    	this.target_checksum = target_checksum;
    }
    
    public SHA1 getTargetChecksum(){
    	return this.target_checksum;
    }
    
    public String getFilepath(){
    	if (this.folder == null || this.folder.isEmpty()) return this.filename;
    	return this.getFolder() + File.separator + this.filename;
    }
    public String getFolder(){
    	return (this.folder == null) ? "" : this.folder;
    }
    public String getFilename(){
    	return (this.filename == null) ? "" : this.filename;
    }
    
    //Generates checksum for the file described by this Descriptor.
    public SHA1 generateChecksum() {
    	return SHA1.getFileHash(this.getFilepath());
    }
    
    // for use in data structures
    @Override
    public int hashCode() {
    	return this.target_checksum.hashCode();
    }

	@Override
	public int compareTo(Object o) {
		return this.target_checksum.compareTo(o);
	}
}
