package tests;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;

import junit.framework.TestCase;
import modules.*;
import types.*;

public class UDPListenerTest extends TestCase {

	private int TIMEOUT = 50;	
	private int recvPort = 4747;
	private int sendPort = 4747;
	private UDPListener listener;
	private DataStorage ds = new DataStorage();
	private HashMap<ModulType, ArrayBlockingQueue<Message> > ms = new HashMap<ModulType, ArrayBlockingQueue<Message>>();
	
	public UDPListenerTest() {
		for (ModulType type: ModulType.values())
			ms.put(type, new ArrayBlockingQueue<Message>(100));
	}
	//--------
	public void testConstructor_invalidPort() {
		try {
			new UDPListener(null, null, 80);
		} catch (IOException e) {
			return;
		} 
		fail();
	}
	
	public void testConstructor_validInput() {
		try {
			UDPListener listener = new UDPListener(null, null, recvPort);
			listener.udpSocket().close();
			if (listener.udpSocket() == null) fail();
			
		} catch (IOException e) {
			fail();
		} 
	}
	//---------
	public void testExit() {
		try {
			UDPListener listener = new UDPListener(null, null, recvPort);
			Thread t = new Thread(listener);
			t.start();

			Thread.sleep(TIMEOUT);
			listener.exit();
			Thread.sleep(TIMEOUT);
			
			if (!listener.udpSocket().isClosed()) fail();
			if (t.isAlive()) fail();
			
		} catch (Exception e) {
			fail();
		}
	}
	//---------
	public void testRecievingMessage_invalidTarget() {
		startListener();
		sendUDPMessage((byte)0x02, true, false, 1024);
		try {
			Message msg =  ms.get(ModulType.SEARCH_ANSWER).poll(100, TimeUnit.MILLISECONDS);
			if (msg != null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testRecievingMessage_invalidDataLength() {
		startListener();
		sendUDPMessage((byte)0x02, false, true, 1024);
		try {
			Message msg =  ms.get(ModulType.SEARCH_ANSWER).poll(100, TimeUnit.MILLISECONDS);
			if (msg != null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testRecievingMessage_notUDPTargetModule() {
		startListener();
		sendUDPMessage((byte)0x05, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.DOWNLOAD).poll(100, TimeUnit.MILLISECONDS);
			if (msg != null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testRecievingMessage_onlyOneModule() {
		startListener();
		sendUDPMessage((byte)0x02, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.SEARCH).poll(100, TimeUnit.MILLISECONDS);
			if (msg != null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testRecievingMessage_shouldRecieve() {
		startListener();
		sendUDPMessage((byte)0x02, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.SEARCH_ANSWER).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	//------
	public void testForwardingMessage_shouldForwardOkMsg() {
		startListener();
		sendUDPMessage((byte)0x02, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.PEER_MANAGEMENT).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testForwardingMessage_shouldForwardInvalidTarget() {
		startListener();
		sendUDPMessage((byte)0x02, true, false, 1024);
		try {
			Message msg = ms.get(ModulType.PEER_MANAGEMENT).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testForwardingMessage_forwardedMsgNotSame() {
		startListener();
		sendUDPMessage((byte)0x02, false, false, 1024);
		try {
			Message msgf = ms.get(ModulType.PEER_MANAGEMENT).poll(100, TimeUnit.MILLISECONDS);
			Message msg = ms.get(ModulType.SEARCH_ANSWER).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null || msgf == null) safefail();
			if (msg == msgf) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testForwardingMessage_invalidDataLength() {
		startListener();
		sendUDPMessage((byte)0x02, false, true, 1024);
		try {
			Message msg = ms.get(ModulType.PEER_MANAGEMENT).poll(100, TimeUnit.MILLISECONDS);
			if (msg != null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testForwardingMessage_notTwice() {
		startListener();
		sendUDPMessage((byte)0x00, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.PEER_MANAGEMENT).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null) safefail();
			msg = ms.get(ModulType.PEER_MANAGEMENT).poll(100, TimeUnit.MILLISECONDS);
			if (msg != null) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	//-------
	public void testValidMessageRecieved_sourceID() {
		startListener();
		sendUDPMessage((byte)0x02, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.SEARCH_ANSWER).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null) safefail();
			if (!ds.myPeerID().equals(msg.sourceID)) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testValidMessageRecieved_targetID() {
		startListener();
		sendUDPMessage((byte)0x02, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.SEARCH_ANSWER).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null) safefail();
			if (!ds.myPeerID().equals(msg.targetID)) safefail();
		} catch (InterruptedException e) {
			safefail();
		}
		stopListener();
	}
	
	public void testValidMessageRecieved_remotePeer() {
		startListener();
		sendUDPMessage((byte)0x02, false, false, 1024);
		try {
			Message msg = ms.get(ModulType.SEARCH_ANSWER).poll(100, TimeUnit.MILLISECONDS);
			if (msg == null) safefail();
			PeerInfo pi = msg.remotePeer;
			if (!ds.myPeerID().equals(pi.id())) safefail();
			if (!InetAddress.getLocalHost().equals(pi.ip())) safefail();
			if (pi.port() != sendPort) safefail();
		} catch (Exception e) {
			safefail();
		}
		stopListener();
	}
	
	//---------- auxiliary
	private void startListener() {
		try {
			Thread.sleep(TIMEOUT);
			listener = new UDPListener(ds, ms, recvPort);
			Thread t = new Thread(listener);
			t.start();		
		}
		catch (Exception e) {
			safefail();
		}
	}
	
	private void stopListener() {
		listener.exit();
	}
	
	private void sendUDPMessage(byte type, boolean invalidTarget, boolean invalidLength, int dataLen) {
		Message msg = new Message();
		msg.messageType = type;
		msg.sourceID = ds.myPeerID();
		msg.targetID = invalidTarget ? new PeerID(ds.myPeerID().value()+1) : ds.myPeerID();
		msg.identifier = new SequenceID(0);
		msg.dataLength = invalidLength ? dataLen + 10 : dataLen;
		msg.data = new byte[dataLen];
		
		byte[] bytes = msg.getBytes();
		
		try {
			DatagramPacket udpPacket = new DatagramPacket(bytes, 0, bytes.length);
			udpPacket.setAddress(InetAddress.getLocalHost());
			udpPacket.setPort(recvPort);
			
			DatagramSocket sock = new DatagramSocket(sendPort);
			sock.send(udpPacket);
			sock.close();
		}
		catch (Exception e){
			safefail();
		}		
	}
	
	private void safefail() {
		stopListener();
		fail();
	}
}
